import logging
from pathlib import Path

import pytest

from fpdf import FPDF, FontFace, HTMLMixin, TextStyle, TitleStyle
from fpdf.drawing import DeviceRGB
from fpdf.errors import FPDFException
from test.conftest import assert_pdf_equal, LOREM_IPSUM, assert_same_file

HERE = Path(__file__).resolve().parent
FONT_DIR = HERE / ".." / "fonts"


def test_html_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()

    initial = 10
    mm_after_image = initial + 300 / pdf.k
    assert round(pdf.get_x()) == 10
    assert round(pdf.get_y()) == 10
    assert round(pdf.w) == 210

    img_path = HERE.parent / "image/png_images/c636287a4d7cb1a36362f7f236564cef.png"
    pdf.write_html(
        f"<center><img src=\"{img_path}\" height='300' width='300'></center>"
    )
    # Unable to text position of the image as write html moves to a new line after
    # adding the image but it can be seen in the resulting html_images.pdf file.
    assert round(pdf.get_x()) == 10
    assert pdf.get_y() == pytest.approx(mm_after_image, abs=0.01)

    assert_pdf_equal(pdf, HERE / "html_images.pdf", tmp_path)


def test_html_features(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("<p><b>hello</b> world. i am <i>tired</i>.</p>")
    pdf.write_html("<p><u><b>hello</b> world. i am <i>tired</i>.</u></p>")
    pdf.write_html("<p><u><strong>hello</strong> world. i am <em>tired</em>.</u></p>")
    pdf.write_html('<p><a href="https://github.com">github</a></p>')
    pdf.write_html('<p align="right">right aligned text</p>')
    pdf.write_html("<p>i am a paragraph <br />in two parts.</p>")
    pdf.write_html('<font color="#00ff00"><p>hello in green</p></font>')
    pdf.write_html('<font size="7"><p>hello small</p></font>')
    pdf.write_html('<font face="helvetica"><p>hello helvetica</p></font>')
    pdf.write_html('<font face="times"><p>hello times</p></font>')
    pdf.write_html("<h1>h1</h1>")
    pdf.write_html("<h2>h2</h2>")
    pdf.write_html("<h3>h3</h3>")
    pdf.write_html("<h4>h4</h4>")
    pdf.write_html("<h5>h5</h5>")
    pdf.write_html("<h6>h6</h6>")
    pdf.write_html("<p>Rendering two &lt;hr&gt; tags:</p>")
    pdf.write_html('<hr style="width: 50%">')
    pdf.write_html("<hr>")
    # Now inserting <br> tags until a page jump is triggered:
    for _ in range(24):
        pdf.write_html("<br>")
    pdf.write_html("<pre>i am preformatted text.</pre>")
    pdf.write_html("<blockquote>hello blockquote</blockquote>")
    pdf.write_html("<ul><li>li1</li><li>another</li><li>l item</li></ul>")
    pdf.write_html("<ol><li>li1</li><li>another</li><li>l item</li></ol>")
    pdf.write_html("<dl><dt>description title</dt><dd>description details</dd></dl>")
    pdf.write_html("<br><br>")
    pdf.write_html(
        "<table>"
        "  <thead>"
        "    <tr>"
        '      <th  width="30%">ID</th>'
        '      <th  width="70%">Name</th>'
        "    </tr>"
        "  </thead>"
        "  <tbody>"
        "    <tr>"
        "      <td>1</td>"
        "      <td>Alice</td>"
        "    </tr>"
        "    <tr>"
        "      <td>2</td>"
        "      <td>Bob</td>"
        "    </tr>"
        "  </tbody>"
        "  <tfoot>"
        "    <tr>"
        "      <td>id</td>"
        "      <td>name</td>"
        "    </tr>"
        "  </tfoot>"
        "</table>"
    )
    pdf.write_html("<br>")
    pdf.write_html(
        '<table width="50%">'
        "  <thead>"
        "    <tr>"
        '      <th  width="30%">ID</th>'
        '      <th  width="70%">Name</th>'
        "    </tr>"
        "  </thead>"
        "  <tbody>"
        "    <tr>"
        "      <td>1</td>"
        "      <td>Alice</td>"
        "    </tr>"
        "    <tr>"
        "      <td>2</td>"
        "      <td>Bob</td>"
        "    </tr>"
        "  </tbody>"
        "  <tfoot>"
        "    <tr>"
        "      <td>id</td>"
        "      <td>name</td>"
        "    </tr>"
        "  </tfoot>"
        "</table>"
    )

    name = [
        "Alice",
        "Carol",
        "Chuck",
        "Craig",
        "Dan",
        "Erin",
        "Eve",
        "Faythe",
        "Frank",
        "Grace",
        "Heidi",
        "Ivan",
        "Judy",
        "Mallory",
        "Michael",
        "Niaj",
        "Olivia",
        "Oscar",
        "Peggy",
        "Rupert",
        "Sybil",
        "Trent",
        "Trudy",
        "Victor",
        "Walter",
        "Wendy",
    ]

    def getrow(i):
        return f"<tr><td>{i}</td><td>{name[i]}</td></tr>"

    pdf.write_html(
        (
            '<table width="50%">'
            "  <thead>"
            "    <tr>"
            '      <th  width="30%">ID</th>'
            '      <th  width="70%">Name</th>'
            "    </tr>"
            "  </thead>"
            "  <tbody>"
            "    <tr>"
            '      <td colspan="2" align="center">Alice</td>'
            "    </tr>"
        )
        + "".join(getrow(i) for i in range(26))
        + "  </tbody>"
        + "</table>"
    )

    pdf.add_page()
    img_path = HERE.parent / "image/png_images/c636287a4d7cb1a36362f7f236564cef.png"
    pdf.write_html(f"<img src=\"{img_path}\" height='300' width='300'>")
    # With an (incorrect) trailing slash:
    pdf.write_html(f"<img src=\"{img_path}\" height='300' width='300'/>")
    # With an (incorrect) end tag:
    pdf.write_html(f"<img src=\"{img_path}\" height='300' width='300'></img>")

    assert_pdf_equal(pdf, HERE / "html_features.pdf", tmp_path)


def test_html_bold_italic_underline(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<B>bold</B>
           <I>italic</I>
           <U>underlined</U>
           <b><i><u>all at once!</u></i></b>""")
    assert_pdf_equal(pdf, HERE / "html_bold_italic_underline.pdf", tmp_path)


def test_html_strikethrough(tmp_path):
    pdf = FPDF()
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("<s>strikethrough</s>")
    pdf.write_html('<font face="DejaVuSans"><s>strikethrough</s></font>')
    assert_pdf_equal(pdf, HERE / "html_strikethrough.pdf", tmp_path)


def test_html_customize_ul(tmp_path):
    html = """<ul>
        <li><b>term1</b>: definition1</li>
        <li><b>term2</b>: definition2</li>
    </ul>"""
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    # Customizing through optional method arguments:
    for indent, bullet in ((5, "\x86"), (10, "\x9b"), (15, "\xac"), (20, "\xb7")):
        pdf.write_html(
            html,
            tag_styles={"li": TextStyle(l_margin=indent, b_margin=2)},
            ul_bullet_char=bullet,
        )
        pdf.ln()
    assert_pdf_equal(pdf, HERE / "html_customize_ul.pdf", tmp_path)


def test_html_customize_ul_deprecated(tmp_path):
    html = """<ul>
        <li><b>term1</b>: definition1</li>
        <li><b>term2</b>: definition2</li>
    </ul>"""
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    with pytest.warns(DeprecationWarning):  # li_tag_indent
        # Customizing through optional method arguments:
        for indent, bullet in ((5, "\x86"), (10, "\x9b"), (15, "\xac"), (20, "\xb7")):
            pdf.write_html(html, li_tag_indent=indent, ul_bullet_char=bullet)
            pdf.ln()
    assert_pdf_equal(pdf, HERE / "html_customize_ul_deprecated.pdf", tmp_path)


def test_html_li_tag_indent_deprecated(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    with pytest.warns(DeprecationWarning):
        pdf.write_html("<ul><li>item 1</li></ul>", li_tag_indent=40)
        pdf.write_html("<ul><li>item 2</li></ul>", li_tag_indent=50)
        pdf.write_html("<ul><li>item 3</li></ul>", li_tag_indent=60)
    assert_pdf_equal(pdf, HERE / "html_li_tag_indent.pdf", tmp_path)


def test_html_ol_start_and_type(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<ol start="2" type="i">
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ol>""")
    assert_pdf_equal(pdf, HERE / "html_ol_start_and_type.pdf", tmp_path)


def test_html_ul_type(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html(text="""
        <ul type="circle">
          <li>a list item</li>
        </ul>
        <ul type="disc">
          <li>another list item</li>
        </ul>""")
    pdf.ln()
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans")
    pdf.write_html("""
        <ul type="■">
          <li>a list item</li>
          <li>another list item</li>
        </ul>""")
    assert_pdf_equal(pdf, HERE / "html_ul_type.pdf", tmp_path)


def test_html_list_with_custom_font(caplog, tmp_path):  # cf. issue #1496
    pdf = FPDF()
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans")
    pdf.add_page()
    pdf.write_html("""<ul>
          <li>item</li>
        </ul>
        <ul type="circle">
          <li>item</li>
        </ul>
        <ul type="disc">
          <li>item</li>
        </ul>""")
    assert_pdf_equal(pdf, HERE / "html_list_with_custom_font.pdf", tmp_path)
    assert "WARN" not in caplog.text


def test_html_li_prefix_color(tmp_path):
    html = """<ul>
        <li>item1</li>
        <li>item2</li>
        <li>item3</li>
    </ul>"""

    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html(html, li_prefix_color=0)  # black
    pdf.ln()
    pdf.write_html(html, li_prefix_color="green")
    pdf.ln()
    pdf.write_html(html, li_prefix_color=DeviceRGB(r=0.5, g=1, b=0))
    pdf.ln()
    assert_pdf_equal(pdf, HERE / "html_li_prefix_color.pdf", tmp_path)


def test_html_align_paragraph(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_margins(50, 20)
    pdf.write_html(f"""
        No align given, default left:
        <p>{LOREM_IPSUM[:200]}"</p>
        align=justify:
        <p align="justify">{LOREM_IPSUM[:200]}"</p>
        align=right:
        <p align="right">{LOREM_IPSUM[200:400]}"</p>
        align=left:
        <p align="left">{LOREM_IPSUM[400:600]}"</p>
        align=center:
        <p align="center">{LOREM_IPSUM[600:800]}"</p>
        <!-- ignore invalid align -->
        align=invalid, ignore and default left:
        <p align="invalid">{LOREM_IPSUM[800:1000]}"</p>""")
    assert_pdf_equal(pdf, HERE / "html_align_paragraph.pdf", tmp_path)


def test_issue_156(tmp_path):
    pdf = FPDF()
    pdf.add_font("Roboto", style="B", fname=FONT_DIR / "Roboto-Bold.ttf")
    pdf.set_font("Roboto", style="B")
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        pdf.write_html("Regular text<br><b>Bold text</b>")
    assert (
        str(error.value)
        == "Undefined font: roboto - Use built-in fonts or FPDF.add_font() beforehand"
    )
    pdf.add_font("Roboto", fname=FONT_DIR / "Roboto-Regular.ttf")
    pdf.write_html("Regular text<br><b>Bold text</b>")
    assert_pdf_equal(pdf, HERE / "issue_156.pdf", tmp_path)


def test_html_font_color_name(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        '<font color="crimson"><p>hello in crimson</p></font>'
        '<font color="#f60"><p>hello in orange</p></font>'
        '<font color="LIGHTBLUE"><p><b>bold hello in light blue</b></p></font>'
        '<font color="royalBlue"><p>hello in royal blue</p></font>'
        '<font color="#000"><p>hello in black</p></font>'
        '<font color="beige"><p><i>italic hello in beige</i></p></font>'
    )
    assert_pdf_equal(pdf, HERE / "html_font_color_name.pdf", tmp_path)


def test_html_heading_hebrew(tmp_path):
    pdf = FPDF()
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans")
    pdf.add_page()
    pdf.write_html("<h1>Hebrew: שלום עולם</h1>")
    assert_pdf_equal(pdf, HERE / "html_heading_hebrew.pdf", tmp_path)


def test_html_headings_line_height(tmp_path):  # issue-223
    pdf = FPDF()
    pdf.add_page()
    long_title = "The Quick Brown Fox Jumped Over The Lazy Dog "
    pdf.write_html(f"""
        <h1>H1   {long_title*2}</h1>
        <h2>H2   {long_title*2}</h2>
        <h3>H3   {long_title*2}</h3>
        <h4>H4   {long_title*3}</h4>
        <h5>H5   {long_title*3}</h5>
        <h6>H6   {long_title*4}</h6>
        <p>P   {long_title*5}</p>""")
    assert_pdf_equal(pdf, HERE / "html_headings_line_height.pdf", tmp_path)


def test_html_custom_heading_sizes(tmp_path):  # issue-223
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """<h1>This is a H1</h1>
        <h2>This is a H2</h2>
        <h3>This is a H3</h3>
        <h4>This is a H4</h4>
        <h5>This is a H5</h5>
        <h6>This is a H6</h6>""",
        tag_styles={
            "h1": TextStyle(
                color="#960000", t_margin=5 + 834 / 900, b_margin=0.4, font_size_pt=6
            ),
            "h2": TextStyle(
                color="#960000", t_margin=5 + 453 / 900, b_margin=0.4, font_size_pt=12
            ),
            "h3": TextStyle(
                color="#960000", t_margin=5 + 199 / 900, b_margin=0.4, font_size_pt=18
            ),
            "h4": TextStyle(
                color="#960000", t_margin=5 + 72 / 900, b_margin=0.4, font_size_pt=24
            ),
            "h5": TextStyle(
                color="#960000", t_margin=5 - 55 / 900, b_margin=0.4, font_size_pt=30
            ),
            "h6": TextStyle(
                color="#960000", t_margin=5 - 182 / 900, b_margin=0.4, font_size_pt=36
            ),
        },
    )
    assert_pdf_equal(pdf, HERE / "html_custom_heading_sizes.pdf", tmp_path)


def test_html_custom_heading_sizes_deprecated(tmp_path):  # issue-223
    pdf = FPDF()
    pdf.add_page()
    with pytest.warns(DeprecationWarning):
        pdf.write_html(
            """<h1>This is a H1</h1>
            <h2>This is a H2</h2>
            <h3>This is a H3</h3>
            <h4>This is a H4</h4>
            <h5>This is a H5</h5>
            <h6>This is a H6</h6>""",
            heading_sizes=dict(h1=6, h2=12, h3=18, h4=24, h5=30, h6=36),
        )
    assert_pdf_equal(pdf, HERE / "html_custom_heading_sizes.pdf", tmp_path)


def test_html_superscript(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        "<h1>Superscript/Subscript test</h1>"
        "2<sup>56</sup> more line text<sub>(idx)</sub>"
    )
    assert_pdf_equal(pdf, HERE / "html_superscript.pdf", tmp_path)


def test_html_description(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""
        <dt>description title</dt>
        <dd>description details</dd>
        <dl>
            <dt>description title</dt>
            <dd>description details</dd>
        </dl>""")
    assert_pdf_equal(pdf, HERE / "html_description.pdf", tmp_path)


def test_html_HTMLMixin_deprecation_warning(tmp_path):
    class PDF(FPDF, HTMLMixin):
        pass

    msg = (
        "The HTMLMixin class is deprecated since v2.6.0. "
        "Simply use the FPDF class as a replacement."
    )

    with pytest.warns(DeprecationWarning, match=msg) as record:
        pdf = PDF()
        pdf.add_page()
        pdf.write_html("""
           <dt>description title</dt>
           <dd>description details</dd>
            <dl>
                <dt>description title</dt>
                <dd>description details</dd>
            </dl>""")
        assert_pdf_equal(pdf, HERE / "html_description.pdf", tmp_path)

    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)


def test_html_whitespace_handling(tmp_path):  # Issue 547
    """Testing whitespace handling for write_html()."""
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""
<body>
<h1>Issue 547 Test</h1>
<p>
<b>Testing   </b> paragraph blocks
        that <i>span</i> <b>multiple lines</b>.
    Testing tabs       and    spaces
    <br>and break tags.
</p>
<code>Testing code blocks with tabs      and    spaces.</code>
<br>
<pre>
Testing pre blocks
that span multiple lines
and have tabs    and    spaces.
</pre>

<pre><code>
Testing pre-code blocks
that span multiple lines
and have tabs    and    spaces.
</code></pre>

<p>Testing unicode nbsp \u00a0\u00a0\u00a0\u00a0,
and html nbsp &nbsp;&nbsp;&nbsp;&nbsp;.
<br>\u00a0&nbsp;&nbsp;Testing leading nbsp
</p>
</body>
""")
    assert_pdf_equal(pdf, HERE / "html_whitespace_handling.pdf", tmp_path)


def test_html_custom_line_height(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<p line-height=3>
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text</p>
<p line-height="2">
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text-</p>
<p line-height="x"><!-- invalid line-height ignored, default value of 1 will be used -->
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text-
text-text-text-text-text-text-text-text-text-text-</p>
""")
    assert_pdf_equal(pdf, HERE / "html_custom_line_height.pdf", tmp_path)


def test_html_img_not_overlapping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """<img src="test/image/png_images/affc57dfffa5ec448a0795738d456018.png"/>
           <p>text</p>"""
    )
    assert_pdf_equal(
        pdf,
        HERE / "html_img_not_overlapping.pdf",
        tmp_path,
    )


def test_html_img_without_height_at_page_bottom_triggers_page_break(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.y = 200
    img_path = HERE.parent / "image/png_images/c636287a4d7cb1a36362f7f236564cef.png"
    pdf.write_html(f'<img width="500" src="{img_path}">')
    assert_pdf_equal(
        pdf,
        HERE / "html_img_without_height_at_page_bottom_triggers_page_break.pdf",
        tmp_path,
    )


def test_warn_on_tags_not_matching(caplog):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("<p>")
    assert "Missing HTML end tag for <p>" in caplog.text
    pdf.write_html("</p>")
    assert " Unexpected HTML end tag </p>" in caplog.text
    pdf.write_html("<p></a>")
    assert " Unexpected HTML end tag </a>" in caplog.text


def test_html_unorthodox_headings_hierarchy(tmp_path):  # issue 631
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<h1>H1</h1>
           <h5>H5</h5>""")
    assert_pdf_equal(pdf, HERE / "html_unorthodox_headings_hierarchy.pdf", tmp_path)


def test_html_custom_pre_code_font(tmp_path):  # issue 770
    pdf = FPDF()
    pdf.add_font(fname=FONT_DIR / "DejaVuSansMono.ttf")
    pdf.add_page()
    pdf.write_html(
        "<code> Cześć! </code>",
        tag_styles={"code": TextStyle(font_family="DejaVuSansMono")},
    )
    assert_pdf_equal(pdf, HERE / "html_custom_pre_code_font.pdf", tmp_path)


def test_html_custom_pre_code_font_deprecated(tmp_path):  # issue 770
    pdf = FPDF()
    pdf.add_font(fname=FONT_DIR / "DejaVuSansMono.ttf")
    pdf.add_page()
    with pytest.warns(DeprecationWarning):
        pdf.write_html("<code> Cześć! </code>", pre_code_font="DejaVuSansMono")
    assert_pdf_equal(pdf, HERE / "html_custom_pre_code_font.pdf", tmp_path)


def test_html_preserve_initial_text_color(tmp_path):  # issue 846
    pdf = FPDF()
    pdf.add_page()
    pdf.set_text_color(200, 50, 50)
    pdf.set_font(family="Helvetica", size=13)
    pdf.write_html("one <font size=8>two</font> three")
    assert_pdf_equal(pdf, HERE / "html_preserve_initial_text_color.pdf", tmp_path)


def test_html_heading_color_attribute(tmp_path):  # discussion 880
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""
        <h1>Title</h1>
        Content
        <h2 color="#00ff00">Subtitle in green</h2>
        Content
        """)
    assert_pdf_equal(pdf, HERE / "html_heading_color_attribute.pdf", tmp_path)


def test_html_format_within_p(tmp_path):  # discussion 880
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("times", size=18)
    pdf.set_margins(20, 20, 100)
    pdf.write_html("""
<p align="justify">This is a sample text that will be justified
in the PDF. <u>This</u> is a <font color="red">sample text</font> that will be justified
in the PDF. <b>This</b> is a sample text that will be justified in the PDF.
<i>This</i> is a sample text that will be justified in the PDF.</p>
        """)
    assert_pdf_equal(pdf, HERE / "html_format_within_p.pdf", tmp_path)


def test_html_bad_font():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("times", size=18)
    with pytest.raises(FPDFException):
        pdf.write_html('<font face="no_such_font"><p>hello helvetica</p></font>')


def test_html_ln_outside_p(tmp_path):
    # Edge case. The <li> must come right after the </p> without anything
    # else in between (which would cause a new paragraph to be started).
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("times", size=18)
    pdf.write_html(
        "<p>something in paragraph</p><li>causing _ln() outside paragraph</li>"
    )
    assert_pdf_equal(pdf, HERE / "html_ln_outside_p.pdf", tmp_path)


def test_html_sections(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    pdf.write_html("""
        <section>
           <h2>Subtitle 1</h2>
            <section>
              <h3>Subtitle 1.1</h3>
              Lorem ipsum dolor sit amet, consectetur adipiscing elit,
              sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
            </section>
            <section>
              <h3>Subtitle 1.2</h3>
              Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
            </section>
        </section>
        """)
    assert_pdf_equal(pdf, HERE / "html_sections.pdf", tmp_path)


def test_html_and_section_title_styles():  # issue 1080
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    pdf.set_section_title_styles(TextStyle("Helvetica", "B", 20, (0, 0, 0)))
    with pytest.raises(NotImplementedError):
        pdf.write_html("""
            <h1>Heading One</h1>
            <p>Just enough text to show how bad the situation really is</p>
            <h2>Heading Two</h2>
            <p>This will not overflow</p>
            """)


def test_html_and_section_title_styles_with_deprecated_TitleStyle():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    with pytest.warns(DeprecationWarning):
        pdf.set_section_title_styles(TitleStyle("Helvetica", "B", 20, (0, 0, 0)))
    with pytest.raises(NotImplementedError):
        pdf.write_html("""
            <h1>Heading One</h1>
            <p>Just enough text to show how bad the situation really is</p>
            <h2>Heading Two</h2>
            <p>This will not overflow</p>
            """)


def test_html_link_underline(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        '<a href="https://py-pdf.github.io/fpdf2/">inside link</a> - outside link'
    )
    pdf.write_html(
        '<a href="https://www.flickr.com/photos/ryzom/14726336322/in/album-72157645935788203/">Tryker</a>'
        " - Ryzom"
        ' - <a href="https://creativecommons.org/licenses/by-sa/2.0/">CC BY-SA 2.0</a>'
    )
    assert_pdf_equal(pdf, HERE / "html_link_underline.pdf", tmp_path)


def test_html_link_style(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = '<a href="http://www.example.com">Link to www.example.com</a>'
    style = FontFace(color="#f00", family="Courier", size_pt=8, emphasis="BIU")
    pdf.write_html(html, tag_styles={"a": style})
    assert_pdf_equal(pdf, HERE / "html_link_style.pdf", tmp_path)


def test_html_blockquote_color(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = "Text before<blockquote>foo</blockquote>Text afterwards"
    blockquote_style = TextStyle(
        color=(125, 125, 0), font_style="ITALICS", t_margin=3, b_margin=3, l_margin=10
    )
    pdf.write_html(html, tag_styles={"blockquote": blockquote_style})
    assert_pdf_equal(pdf, HERE / "html_blockquote_color.pdf", tmp_path)


def test_html_headings_color(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = "<h1>foo</h1><h2>bar</h2>"
    pdf.write_html(
        html,
        tag_styles={
            "h1": TextStyle(
                color=(148, 139, 139),
                font_size_pt=24,
                t_margin=5 + 834 / 900,
                b_margin=0.4,
            ),
            "h2": TextStyle(
                color=(148, 139, 139),
                font_size_pt=18,
                t_margin=5 + 453 / 900,
                b_margin=0.4,
            ),
        },
    )
    assert_pdf_equal(pdf, HERE / "html_headings_color.pdf", tmp_path)


def test_html_unsupported_tag_color():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(NotImplementedError):
        pdf.write_html("<p>foo</p><hr><p>bar</p>", tag_styles={"hr": TextStyle()})


def test_html_link_style_using_TextStyle(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = '<a href="http://www.example.com">Link to www.example.com</a>'
    style = TextStyle(
        color="#f00", font_family="Courier", font_size_pt=8, font_style="BIU"
    )
    pdf.write_html(html, tag_styles={"a": style})
    assert_pdf_equal(pdf, HERE / "html_link_style.pdf", tmp_path)


def test_html_blockquote_color_using_FontFace(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = "Text before<blockquote>foo</blockquote>Text afterwards"
    pdf.write_html(
        html,
        tag_styles={"blockquote": FontFace(color=(125, 125, 0), emphasis="ITALICS")},
    )
    assert_pdf_equal(pdf, HERE / "html_blockquote_color_using_FontFace.pdf", tmp_path)


def test_html_headings_color_using_FontFace(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = "<h1>foo</h1><h2>bar</h2>"
    pdf.write_html(
        html,
        tag_styles={
            "h1": FontFace(color=(148, 139, 139), size_pt=24),
            "h2": FontFace(color=(148, 139, 139), size_pt=18),
        },
    )
    assert_pdf_equal(pdf, HERE / "html_headings_color.pdf", tmp_path)


def test_html_unsupported_tag_color_using_FontFace():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(NotImplementedError):
        pdf.write_html("<p>foo</p><hr><p>bar</p>", tag_styles={"hr": FontFace()})


def test_html_blockquote_indent(tmp_path):  # issue-1074
    pdf = FPDF()
    pdf.add_page()
    html = "Text before<blockquote>foo</blockquote>Text afterwards"
    pdf.write_html(
        html,
        tag_styles={
            "blockquote": TextStyle(
                color=(100, 0, 45), t_margin=3, b_margin=3, l_margin=20
            )
        },
    )
    html = (
        "<blockquote>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod"
        "tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,"
        "quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."
        "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu"
        "fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident,"
        "sunt in culpa qui officia deserunt mollit anim id est laborum.</blockquote>"
    )
    pdf.write_html(
        html,
        tag_styles={
            "blockquote": TextStyle(
                color=(100, 0, 45), t_margin=3, b_margin=3, l_margin=40
            )
        },
    )
    assert_pdf_equal(pdf, HERE / "html_blockquote_indent.pdf", tmp_path)


def test_html_blockquote_indent_using_deprecated_tag_indents(tmp_path):  # issue-1074
    pdf = FPDF()
    pdf.add_page()
    html = "Text before<blockquote>foo</blockquote>Text afterwards"
    with pytest.warns(DeprecationWarning):
        pdf.write_html(html, tag_indents={"blockquote": 20})
    html = (
        "<blockquote>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod"
        "tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,"
        "quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."
        "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu"
        "fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident,"
        "sunt in culpa qui officia deserunt mollit anim id est laborum.</blockquote>"
    )
    with pytest.warns(DeprecationWarning):
        pdf.write_html(html, tag_indents={"blockquote": 40})
    assert_pdf_equal(pdf, HERE / "html_blockquote_indent.pdf", tmp_path)


def test_html_ol_ul_line_height(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<p>Default line-height:</p>
        <ul>
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ul>
        <p>1.5 line-height:</p>
        <ol line-height="1.5">
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ol>
        <p>Double line-height:</p>
        <ul line-height="2">
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ul>
        <p>1.5 line-height as "style":</p>
        <ol style="line-height: 1.5">
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ol>
        <p>Double line-height as "style":</p>
        <ul style="line-height: 2">
            <li>item</li>
            <li>item</li>
            <li>item</li>
        </ul>""")
    assert_pdf_equal(pdf, HERE / "html_ol_ul_line_height.pdf", tmp_path)


def test_html_long_list_entries(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html = f"<ul><li>{'A ' * 200}</li></ul>"
    pdf.write_html(html)
    assert_pdf_equal(pdf, HERE / "html_long_list_entries.pdf", tmp_path)


def test_html_long_ol_bullets(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html_arabic_indian = f"""
        <ol start="{10**100}">
            <li>Item 1</li>
            <li>Item 2</li>
            <li>Item 3</li>
        </ol>"""
    pdf.write_html(html_arabic_indian)
    html_roman = f"""
        <ol start="{10**5}" type="i">
            <li>Item 1</li>
            <li>Item 2</li>
            <li>Item 3</li>
        </ol>"""
    pdf.write_html(html_roman)
    pdf.write_html(
        html_arabic_indian, tag_styles={"li": TextStyle(l_margin=50, t_margin=2)}
    )
    pdf.write_html(html_roman, tag_styles={"li": TextStyle(l_margin=100, t_margin=2)})
    assert_pdf_equal(pdf, HERE / "html_long_ol_bullets.pdf", tmp_path)


def test_html_long_ol_bullets_deprecated(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    html_arabic_indian = f"""
        <ol start="{10**100}">
            <li>Item 1</li>
            <li>Item 2</li>
            <li>Item 3</li>
        </ol>"""
    pdf.write_html(html_arabic_indian)
    html_roman = f"""
        <ol start="{10**5}" type="i">
            <li>Item 1</li>
            <li>Item 2</li>
            <li>Item 3</li>
        </ol>"""
    pdf.write_html(html_roman)
    with pytest.warns(DeprecationWarning):
        pdf.write_html(html_arabic_indian, tag_indents={"li": 50})
        pdf.write_html(html_roman, tag_indents={"li": 100})
    assert_pdf_equal(pdf, HERE / "html_long_ol_bullets.pdf", tmp_path)


def test_html_measurement_units(tmp_path):
    for unit in ["pt", "mm", "cm", "in"]:
        pdf = FPDF(unit=unit)
        pdf.add_page()
        html = """
            <ul>
                <li>Item 1</li>
                <li>Item 2</li>
                <li>Item 3</li>
            </ul>
            <ol>
                <li>Item 1</li>
                <li>Item 2</li>
                <li>Item 3</li>
            </ol>
            <blockquote>Blockquote text</blockquote>
            <dt>Description title</dt>
            <dd>Description details</dd>"""
        pdf.write_html(html)
        assert_pdf_equal(pdf, HERE / "html_measurement_units.pdf", tmp_path)


def test_bulleted_paragraphs():
    pdf = FPDF()
    pdf.add_page()
    text_columns = pdf.text_columns(skip_leading_spaces=True)
    cases = [
        {"indent": 1, "bullet_string": None},
        {
            "indent": 10,
            "bullet_string": "",
            "bullet_r_margin": 2,
        },
        {
            "indent": 20,
            "bullet_string": "a",
            "bullet_r_margin": 0,
        },
        {
            "indent": -20,
            "bullet_string": "abcd",
            "bullet_r_margin": 4,
        },
        {
            "indent": 1000,
            "bullet_string": "abcd\nfghi",
            "bullet_r_margin": -3,
        },
    ]
    pdf.set_font("helvetica", style="B", size=16)
    for case in cases:
        try:
            text_columns.paragraph(
                indent=case.get("indent"),
                bullet_string=case.get("bullet_string"),
                bullet_r_margin=case.get("bullet_r_margin"),
            )
            text_columns.end_paragraph()
        except FPDFException as error:
            pytest.fail(
                f"case: (indent: {case['indent']}, bullet_string: {case['bullet']})\n"
                + str(error)
            )
    bad_bullet_string = "我"
    with pytest.raises(FPDFException) as error:
        text_columns.paragraph(indent=1, bullet_string=bad_bullet_string)
    expected_msg = (
        f'Character "{bad_bullet_string}" at index {0} in text is outside the range of characters '
        f'supported by the font used: "{pdf.font_family+pdf.font_style}". Please consider using a Unicode font.'
    )
    assert str(error.value) == expected_msg


def test_html_list_vertical_margin(tmp_path):
    pdf = FPDF()
    for margin_value in (None, 4, 8, 16):
        pdf.add_page()
        html = f"""
            This page uses `t_margin={margin_value}` for &lt;ul&gt; tags:
            <ul>
                <li>Item 1</li>
                <li>Item 2</li>
                <li>Item 3</li>
            </ul>
            <ol>
                <li>Item 1</li>
                <li>Item 2</li>
                <li>Item 3</li>
            </ol>"""
        pdf.write_html(
            html,
            tag_styles={
                "ol": TextStyle(t_margin=margin_value),
                "ul": TextStyle(t_margin=margin_value),
            },
        )
    assert_pdf_equal(pdf, HERE / "html_list_vertical_margin.pdf", tmp_path)


def test_html_page_break_before(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""Content on first page.
        <br style="break-before: page">
        Content on second page, with some slight top margin.
        <p style="break-before: page">
        Content on third page.
        </p>""")
    assert_pdf_equal(pdf, HERE / "html_page_break_before.pdf", tmp_path)


def test_html_page_break_after(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""Content on first page.
        <br style="break-after: page">
        Content on second page.
        <p style="break-after: page">
        Other content on second page.
        </p>
        Content on third page.""")
    assert_pdf_equal(pdf, HERE / "html_page_break_after.pdf", tmp_path)


def test_html_heading_above_below(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """
        <h1>Top heading</h1>
        <p>Lorem ipsum</p>
        <h2>First heading</h2>
        <p>Lorem ipsum</p>
        <h2>Second heading</h2>
        <p>Lorem ipsum</p>""",
        tag_styles={
            "h1": TextStyle(
                color="#960000", t_margin=10, b_margin=0.5, font_size_pt=24
            ),
            "h2": TextStyle(
                color="#960000", t_margin=10, b_margin=0.5, font_size_pt=18
            ),
        },
    )
    assert_pdf_equal(pdf, HERE / "html_heading_above_below.pdf", tmp_path)


def test_html_dd_tag_indent_deprecated(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        "<dl><dt>description title</dt><dd>description details</dd></dl>",
        tag_styles={"dd": TextStyle(l_margin=5)},
    )
    assert_pdf_equal(pdf, HERE / "html_dd_tag_indent_deprecated.pdf", tmp_path)
    pdf = FPDF()
    pdf.add_page()
    with pytest.warns(DeprecationWarning):
        pdf.write_html(
            "<dl><dt>description title</dt><dd>description details</dd></dl>",
            dd_tag_indent=5,
        )
    assert_pdf_equal(pdf, HERE / "html_dd_tag_indent_deprecated.pdf", tmp_path)


def test_html_font_family(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        "<p><b>hello</b> world. i am <i>sleepy</i>.</p>",
        font_family="Helvetica",
    )
    assert_pdf_equal(pdf, HERE / "html_font_family.pdf", tmp_path)


def test_html_footer_with_call_to_write_html_ko(tmp_path):  # issue-1222
    "This used to cause a RecursionError"

    class MyPDF(FPDF):
        def footer(self):
            self.set_y(-15)
            self.write_html("<p>Footer</p>")

    pdf = MyPDF()
    pdf.add_page()
    pdf.write_html("<p>Main content</p>")
    assert_pdf_equal(pdf, HERE / "html_footer_with_call_to_write_html_ko.pdf", tmp_path)


def test_html_footer_with_call_to_write_html_ok(tmp_path):  # issue-1222
    class MyPDF(FPDF):
        def footer(self):
            self.set_y(-30)
            self.write_html("<p>Footer</p>")

    pdf = MyPDF()
    pdf.add_page()
    pdf.write_html("<p>Main content</p>")
    assert_pdf_equal(pdf, HERE / "html_footer_with_call_to_write_html_ok.pdf", tmp_path)


def test_html_font_tag(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        '<font size="36">Large text in Times 1<p>Large text in Times 2</p></font>',
    )
    pdf.write_html("<br><hr><br>")
    pdf.write_html(
        """Text in Times 1
        <font face="helvetica">
            Text in Helvetica 1
            <font size="36">
                Large text in Helvetica 2
                <font face="times">
                Large text in Times 2
                <p>Large text in Times 3</p>
                </font>
            Large text in Helvetica 3
            <p>Large text in Helvetica 4</p>
            </font>
            Text in Helvetica 5
        </font>
        Text in Times 4""",
    )
    assert_pdf_equal(pdf, HERE / "html_font_tag.pdf", tmp_path)


def test_html_title(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<head>
            <title>Document title</title>
        </head>""")
    assert_pdf_equal(pdf, HERE / "html_title.pdf", tmp_path)


def test_html_title_with_render_title_tag(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """<head>
            <title>Document title</title>
        </head>""",
        render_title_tag=True,
    )
    assert_pdf_equal(pdf, HERE / "html_title_with_render_title_tag.pdf", tmp_path)


def test_html_title_in_body(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<body>
            <title>Document title</title>
        </body>""")
    assert_pdf_equal(pdf, HERE / "html_title_in_body.pdf", tmp_path)


def test_html_title_duplicated(caplog, tmp_path):
    pdf = FPDF()
    pdf.add_page()
    with caplog.at_level(logging.WARN):
        pdf.write_html("""<head>
                <title>Hello</title>
                <title>World</title>
            </head>""")
    assert 'Ignoring repeated <title> "World"' in caplog.text
    assert_pdf_equal(pdf, HERE / "html_title_duplicated.pdf", tmp_path)


def test_html_ol_nested_in_ul(tmp_path):  # cf. issue #1358
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<ul>
          <li>item
          <ol>
            <li>sub-item</li>
          </ol>
          </li>
        </ul>""")
    assert_pdf_equal(pdf, HERE / "html_ol_nested_in_ul.pdf", tmp_path)



from fpdf import FPDF


class PDF(FPDF):
    def header(self):
        # Rendering logo:
        self.image("../docs/fpdf2-logo.png", 10, 8, 33)
        # Setting font: helvetica bold 15
        self.set_font("helvetica", style="B", size=15)
        # Moving cursor to the right:
        self.cell(80)
        # Printing title:
        self.cell(30, 10, "Title", border=1, align="C")
        # Performing a line break:
        self.ln(20)

    def footer(self):
        # Position cursor at 1.5 cm from bottom:
        self.set_y(-15)
        # Setting font: helvetica italic 8
        self.set_font("helvetica", style="I", size=8)
        # Printing page number:
        self.cell(0, 10, f"Page {self.page_no()}/{{nb}}", align="C")


# Instantiation of inherited class
pdf = PDF()
pdf.add_page()
pdf.set_font("Times", size=12)
for i in range(1, 41):
    pdf.cell(0, 10, f"Printing line number {i}", new_x="LMARGIN", new_y="NEXT")
pdf.output("new-tuto2.pdf")



# test/test_text_rendering.py
import math
from pathlib import Path
import pytest
from fpdf import FPDF
from fpdf.drawing import GradientPaint, GraphicsContext, PaintedPath
from fpdf.drawing_primitives import DeviceRGB, Transform
from fpdf.drawing import Text, TextRun
from fpdf.pattern import LinearGradient
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def _new_pdf():
    pdf = FPDF(unit="pt", format="A4")
    pdf.add_page()
    return pdf


def _rect(x, y, w, h):
    return PaintedPath().rectangle(x, y, w, h)


def test_text_and_shape_order(tmp_path: Path):
    pdf = _new_pdf()

    gc = GraphicsContext()

    # Text first
    t = Text(
        x=40,
        y=80,
        text_runs=(
            TextRun(
                text="Hello Overlap",
                family="helvetica",
                emphasis="",
                size=24,
            ),
        ),
        text_anchor="start",
    )
    gc.add_item(t)

    # Opaque rectangle drawn AFTER text (covers part of it)
    r = _rect(60, 55, 120, 30)
    r.style.fill_color = "#FF0000"
    r.style.stroke_color = "#000000"
    r.style.stroke_width = 0.5
    gc.add_item(r)

    with pdf.drawing_context() as dc:
        dc.add_item(gc)

    assert_pdf_equal(pdf, HERE / "generated_pdf" / "text_shape_order.pdf", tmp_path)


def test_shape_with_transparency_over_text(tmp_path: Path):
    pdf = _new_pdf()

    gc = GraphicsContext()

    # Text baseline
    t = Text(
        x=40,
        y=130,
        text_runs=(
            TextRun(
                text="Transparent Shape Over Me",
                family="helvetica",
                emphasis="",
                size=26,
            ),
        ),
    )
    gc.add_item(t)

    # Semi-transparent blue rectangle over the text
    rect = _rect(35, 105, 260, 32)
    rect.style.fill_color = DeviceRGB(r=0, g=0, b=1, a=0.35)
    rect.style.stroke_color = None
    gc.add_item(rect)

    with pdf.drawing_context() as dc:
        dc.add_item(gc)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "shape_alpha_over_text.pdf", tmp_path
    )


def test_text_with_transparency_over_shape(tmp_path: Path):
    pdf = _new_pdf()

    root = GraphicsContext()

    # Background shape (opaque)
    bg = _rect(30, 180, 310, 50)
    bg.style.fill_color = "#EEEEEE"
    bg.style.stroke_color = "#000000"
    bg.style.stroke_width = 0.5
    root.add_item(bg)

    # Child GC that carries alpha in its style for the text fill
    gc_text = GraphicsContext()
    gc_text.style.fill_color = DeviceRGB(r=0, g=0, b=0, a=0.4)

    t = Text(
        x=40,
        y=210,
        text_runs=(
            TextRun(
                text="Semi-Transparent Text",
                family="helvetica",
                emphasis="",
                size=28,
            ),
        ),
    )
    gc_text.add_item(t)
    root.add_item(gc_text)

    with pdf.drawing_context() as dc:
        dc.add_item(root)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "text_alpha_over_shape.pdf", tmp_path
    )


def test_text_gradient_transparency(tmp_path: Path):
    pdf = _new_pdf()

    square = PaintedPath()
    square.rectangle(30, 250, 250, 50)
    square.style.fill_color = "#000000"

    gradient = LinearGradient(
        from_x=30, from_y=0, to_x=280, to_y=0, colors=["#ff0000", "#ffff00"]
    )
    text = PaintedPath()
    text.add_path_element(
        Text(
            x=40,
            y=280,
            text_runs=(
                TextRun(
                    text="Gradient Text Fill",
                    family="helvetica",
                    emphasis="",
                    size=30,
                ),
            ),
        )
    )
    text.style.fill_color = GradientPaint(gradient, units="userSpaceOnUse")

    with pdf.drawing_context() as dc:
        dc.add_item(square)
        dc.add_item(text)

    assert_pdf_equal(pdf, HERE / "generated_pdf" / "text_fill_gradient.pdf", tmp_path)


def test_text_fonts_styles_alignment(tmp_path: Path):
    pdf = _new_pdf()

    root = GraphicsContext()

    x = 300  # common anchor for alignment tests

    # Visual guide line
    guide = PaintedPath().move_to(x, 20).line_to(x, 320)
    guide.style.stroke_color = "#00AA00"
    guide.style.stroke_width = 0.4
    root.add_item(guide)

    items = [
        Text(
            x=x,
            y=80,
            text_runs=(
                TextRun(
                    text="Start / Helvetica",
                    family="DejaVu, sans-serif",
                    emphasis="",
                    size=20,
                ),
            ),
            text_anchor="start",
        ),
        Text(
            x=x,
            y=110,
            text_runs=(
                TextRun(
                    text="Middle / Helvetica B",
                    family="non-existent, helvetica",
                    emphasis="B",
                    size=20,
                ),
            ),
            text_anchor="middle",
        ),
        Text(
            x=x,
            y=140,
            text_runs=(
                TextRun(
                    text="End / Helvetica I",
                    family="helvetica",
                    emphasis="I",
                    size=20,
                ),
            ),
            text_anchor="end",
        ),
        Text(
            x=x,
            y=170,
            text_runs=(
                TextRun(
                    text="Start / Times BI",
                    family="times",
                    emphasis="BI",
                    size=20,
                ),
            ),
            text_anchor="start",
        ),
        Text(
            x=x,
            y=200,
            text_runs=(
                TextRun(
                    text="Middle / Courier",
                    family="courier",
                    emphasis="",
                    size=20,
                ),
            ),
            text_anchor="middle",
        ),
        Text(
            x=x,
            y=230,
            text_runs=(
                TextRun(
                    text="End / Courier B",
                    family="courier",
                    emphasis="B",
                    size=20,
                ),
            ),
            text_anchor="end",
        ),
    ]
    for it in items:
        root.add_item(it)

    with pdf.drawing_context() as dc:
        dc.add_item(root)

    assert_pdf_equal(
        pdf, HERE / "generated_pdf" / "text_fonts_styles_alignment.pdf", tmp_path
    )


def test_text_font_not_found_raises():
    pdf = _new_pdf()

    gc = GraphicsContext()

    t = Text(
        x=50,
        y=320,
        text_runs=(
            TextRun(
                text="Should fail to resolve font",
                family="__no_such_font_family__",
                emphasis="",
                size=18,
            ),
        ),
    )
    gc.add_item(t)

    with pytest.raises(KeyError) as err:
        with pdf.drawing_context() as dc:
            dc.add_item(gc)
    assert "No suitable font for family='__no_such_font_family__'" in str(err.value)


def test_text_ttf_transforms(tmp_path: Path):
    pdf = _new_pdf()

    pdf.add_font(
        family="DejaVu",
        fname=HERE.parent / "fonts" / "DejaVuSans.ttf",
    )
    pdf.add_font(
        family="NotoEmoji",
        fname=HERE.parent / "color_font" / "colrv1-NotoColorEmoji.ttf",
    )

    root = GraphicsContext()

    # Untransformed guide & DejaVu lines (baseline reference)
    x = 300
    guide = PaintedPath().move_to(x, 30).line_to(x, 330)
    guide.style.stroke_color = "#808080"
    guide.style.stroke_width = 0.4
    root.add_item(guide)

    root.add_item(
        Text(
            x=x,
            y=70,
            text_runs=(
                TextRun(
                    text="Start / DejaVu",
                    family="DejaVu",
                    emphasis="",
                    size=20,
                ),
            ),
            text_anchor="start",
        )
    )
    root.add_item(
        Text(
            x=x,
            y=100,
            text_runs=(
                TextRun(
                    text="Middle / DejaVu",
                    family="DejaVu",
                    emphasis="",
                    size=20,
                ),
            ),
            text_anchor="middle",
        )
    )
    root.add_item(
        Text(
            x=x,
            y=130,
            text_runs=(
                TextRun(
                    text="End / DejaVu",
                    family="DejaVu",
                    emphasis="",
                    size=20,
                ),
            ),
            text_anchor="end",
        )
    )

    # A rotated/translated child GC to test transforms with text anchors
    # Rotate around (x, 220) by -25 degrees
    gc_rot = GraphicsContext()
    gc_rot.transform = Transform.rotation(math.radians(-25)).about(x, 220)

    # A green guide line inside the rotated GC (to visualize the new anchor line)
    guide_rot = PaintedPath().move_to(x, 160).line_to(x, 280)
    guide_rot.style.stroke_color = "#00AA00"
    guide_rot.style.stroke_width = 0.4
    gc_rot.add_item(guide_rot)

    # DejaVu under transform (anchors should align on the rotated green guide)
    gc_rot.add_item(
        Text(
            x=x,
            y=190,
            text_runs=(
                TextRun(
                    text="Start / DejaVu (rot)",
                    family="dejavu, helvetica, sans serif",
                    emphasis="",
                    size=18,
                ),
            ),
            text_anchor="start",
        )
    )
    gc_rot.add_item(
        Text(
            x=x,
            y=220,
            text_runs=(
                TextRun(
                    text="Middle / DejaVu (rot)",
                    family="DejaVu",
                    emphasis="",
                    size=18,
                ),
            ),
            text_anchor="middle",
        )
    )
    gc_rot.add_item(
        Text(
            x=x,
            y=250,
            text_runs=(
                TextRun(
                    text="End / DejaVu (rot)",
                    family="DEJAVU",
                    emphasis="",
                    size=18,
                ),
            ),
            text_anchor="end",
        )
    )

    # NotoEmoji sample (TTF shaping/encoding path) – not rotated
    root.add_item(
        Text(
            x=40,
            y=310,
            text_runs=(
                TextRun(
                    text="😀🚀☕",
                    family="NotoEmoji",
                    emphasis="",
                    size=28,
                ),
            ),
        )
    )

    root.add_item(gc_rot)

    with pdf.drawing_context() as dc:
        dc.add_item(root)

    assert_pdf_equal(pdf, HERE / "generated_pdf" / "text_ttf_transforms.pdf", tmp_path)



import fpdf
import pytest


def test_rect_transform_width_exception():
    svg_data = """<?xml version="1.0" standalone="no"?>
    <svg width="10%" height="10cm" viewBox="0 0 1000 1000" xmlns="http://www.w3.org/2000/svg" version="1.1">
    </svg>
    """

    svg = fpdf.svg.SVGObject(svg_data)
    with pytest.raises(ValueError) as error:
        svg.transform_to_rect_viewport(scale=1, width=None, height=10)
    assert (
        str(error.value)
        == 'SVG "width" is a percentage, hence a viewport width is required'
    )


def test_rect_transform_height_exception():
    svg_data = """<?xml version="1.0" standalone="no"?>
    <svg width="10cm" height="10%" viewBox="0 0 1000 1000" xmlns="http://www.w3.org/2000/svg" version="1.1">
    </svg>
    """

    svg = fpdf.svg.SVGObject(svg_data)
    with pytest.raises(ValueError) as error:
        svg.transform_to_rect_viewport(scale=1, width=10, height=None)
    assert (
        str(error.value)
        == 'SVG "height" is a percentage, hence a viewport height is required'
    )


def test_rect_transform_vw_zero():
    svg_data = """<?xml version="1.0" standalone="no"?>
    <svg width="10cm" height="10cm" viewBox="0 0 0 1000" xmlns="http://www.w3.org/2000/svg" version="1.1">
    </svg>
    """

    svg = fpdf.svg.SVGObject(svg_data)
    width, height, context = svg.transform_to_rect_viewport(scale=2, width=10, height=5)
    assert width == 0
    assert height == 0
    assert context is not None



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_transparency(tmp_path):
    "Inspired by https://github.com/digidigital/Extensions-and-Scripts-for-pyFPDF-fpdf2/tree/main/alpha"
    pdf = FPDF()
    pdf.set_font("Helvetica", style="B", size=24)
    pdf.set_line_width(1.5)
    pdf.add_page()

    # Draw an opaque red square:
    pdf.set_fill_color(255, 0, 0)
    pdf.rect(10, 10, 40, 40, "DF")

    # Set alpha to semi-transparency for shape lines & filled areas:
    with pdf.local_context(fill_opacity=0.5, stroke_opacity=0.5):
        # Draw a green square:
        pdf.set_fill_color(0, 255, 0)
        pdf.rect(20, 20, 40, 40, "DF")

    # Set transparency for images & text:
    with pdf.local_context(fill_opacity=0.25):
        # Insert an image:
        pdf.image(HERE / "../docs/fpdf2-logo.png", 30, 30, 40)
        # Print some text:
        pdf.text(22, 29, "You are...")

    # Print some text with full opacity:
    pdf.text(30, 45, "Over the top")

    assert_pdf_equal(pdf, HERE / "transparency.pdf", tmp_path)



from pathlib import Path
from pytest import warns

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def next_row(pdf):
    pdf.ln()
    pdf.set_y(pdf.get_y() + size + margin)


size = 50
margin = 10


def test_line(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    def draw_diagonal_line(pdf, x, y):
        pdf.line(x, y, x + size, y + size / 2)

    for width in [0.71, 1, 2]:
        pdf.set_line_width(width)
        draw_diagonal_line(pdf, pdf.get_x(), pdf.get_y())
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    for color in [70, 140, 200]:
        pdf.set_draw_color(color)
        draw_diagonal_line(pdf, pdf.get_x(), pdf.get_y())
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_line.pdf", tmp_path)


def test_dash(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    def draw_diagonal_dash(pdf, x, y, *a, **k):
        with warns(DeprecationWarning):
            pdf.dashed_line(x, y, x + size, y + size / 2, *a, **k)

    for width in [0.71, 1, 2]:
        pdf.set_line_width(width)
        draw_diagonal_dash(pdf, pdf.get_x(), pdf.get_y(), margin, margin / 2)
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    for color in [70, 140, 200]:
        pdf.set_draw_color(color)
        draw_diagonal_dash(pdf, pdf.get_x(), pdf.get_y(), margin, margin / 2)
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    pdf.set_draw_color(0)
    pdf.set_line_width(0.2)
    draw_diagonal_dash(pdf, pdf.get_x(), pdf.get_y(), margin, margin)

    pdf.set_x(pdf.get_x() + size + margin)
    draw_diagonal_dash(pdf, pdf.get_x(), pdf.get_y(), margin / 2, margin)

    next_row(pdf)
    pdf.set_line_width(1)
    with warns(DeprecationWarning):
        x, y = pdf.get_x(), pdf.get_y()
        pdf.dashed_line(x, y, x + 100, y + 80, 10, 3)
        pdf.set_x(pdf.get_x() + 20)
        x, y = pdf.get_x(), pdf.get_y()
        pdf.dashed_line(x, y, x + 100, y + 80, 3, 20)
        pdf.set_x(pdf.get_x() + 20)
        x, y = pdf.get_x(), pdf.get_y()
        pdf.dashed_line(x, y, x + 100, y + 80, 6, 17)

    assert_pdf_equal(pdf, HERE / "class_dash.pdf", tmp_path)



from pathlib import Path

import pytest

from fpdf import FPDF, FPDFException
from fpdf.drawing import DeviceCMYK
from test.conftest import assert_pdf_equal

TEXT_SIZE, SPACING = 36, 1.15
LINE_HEIGHT = TEXT_SIZE * SPACING


HERE = Path(__file__).resolve().parent


def test_text_positioning(tmp_path):
    doc = FPDF()
    doc.add_page()
    for i in range(15):  # core font
        doc.set_font("helvetica", size=10 + 2 * i)
        x = 20
        y = 15 + 15 * i
        doc.line(x - 5, y, x + 5, y)
        doc.line(x, y - 5, x, y + 5)
        doc.text(x, y, f"{doc.font_family} {doc.font_size_pt}")
    doc.add_font("Roboto", fname=HERE / "../fonts/Roboto-Regular.ttf")
    for i in range(15):  # unicode font
        doc.set_font("Roboto", size=10 + 2 * i)
        x = 120
        y = 15 + 15 * i
        doc.line(x - 5, y, x + 5, y)
        doc.line(x, y - 5, x, y + 5)
        doc.text(x, y, f"{doc.font_family} {doc.font_size_pt}")

    assert_pdf_equal(doc, HERE / "text_positioning.pdf", tmp_path)


def test_text_stretch_spacing(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=20)
    for i, (stretch, spacing) in enumerate(
        (
            (100, 0),
            (150, 0),
            (100, 10),
            (130, 5),
            (70, 0),
            (100, -5),
            (80, -2),
        )
    ):
        doc.set_stretching(stretch)
        doc.set_char_spacing(spacing)
        doc.text(20, 20 + 10 * i, f"stretch: {stretch}; spacing: {spacing}")
    assert_pdf_equal(doc, HERE / "text_stretch_spacing.pdf", tmp_path)


def test_text_text_mode(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=40)
    doc.set_draw_color(255, 100, 0)
    doc.line_width = 2
    for i, mode in enumerate(
        (
            "FILL",
            "STROKE",
            "FILL_STROKE",
        )
    ):
        doc.text_mode = mode
        doc.text(20, 20 + 20 * i, f"Text mode: {mode}")
    assert_pdf_equal(doc, HERE / "text_text_mode.pdf", tmp_path)


def test_text_color(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=20)
    for i, color in enumerate(
        (
            (255, 100, 100),
            (255, 255, 100),
            (255, 100, 255),
            (250, 250, 250),
            (0, 0, 0),
        )
    ):
        doc.set_text_color(*color)
        doc.text(20, 20 + 10 * i, f"{color}")
    color = DeviceCMYK(0.32, 0, 0.1, 0.59)
    doc.set_text_color(color)
    i += 1
    doc.text(20, 20 + 10 * i, f"{color}")
    assert_pdf_equal(doc, HERE / "text_color.pdf", tmp_path)


def test_text_no_font_set():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        pdf.text(20, 20, text="Hello World!")
    expected_msg = "No font set, you need to call set_font() beforehand"
    assert str(error.value) == expected_msg


def test_text_badinput():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(TypeError):
        pdf.text("x", 20, "Hello World")
    with pytest.raises(TypeError):
        pdf.text(20, "y", "Hello World")
    with pytest.raises(AttributeError):
        pdf.text(20, 20, 777)
    with pytest.raises(AttributeError):
        pdf.text(20, 20, (1, 2, 3))
    with pytest.raises(AttributeError):
        pdf.text(20, 20, None)


def test_text_deprecated_txt_arg():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=10)
    with pytest.warns(
        DeprecationWarning, match='The parameter "txt" has been renamed to "text"'
    ):
        # pylint: disable=unexpected-keyword-arg
        pdf.text(20, 20, txt="Lorem ipsum Ut nostrud irure")



#!/usr/bin/env python3

# Generate a HTML page that makes it easy to visually compare all PDF files
# that are modified in the current branch, compared to the master branch.

# USAGE: ./compare-changed-pdfs.py [test_subdir_path]

import sys, webbrowser
from functools import partial
from http.server import HTTPServer, SimpleHTTPRequestHandler
from os import scandir
from pathlib import Path
from subprocess import check_output

from jinja2 import Environment, FileSystemLoader

PORT = 8000
TEMPLATE_FILENAME = "changed_pdfs_comparison.html"

SCRIPTS_DIR = Path(__file__).parent
REPO_DIR = SCRIPTS_DIR.parent
GIT_REF = "master"
TMP_DIR = REPO_DIR / f"{GIT_REF}-checkouts"
VIEW_SLICE = slice(0, 50)


def scantree_dirs(path):
    "Recursively yield DirEntry objects for all sub-directories in the given folder"
    yield path
    for entry in scandir(path):
        if entry.is_dir():
            yield from scantree_dirs(entry.path)


target_dir = sys.argv[1] if len(sys.argv) > 1 else "test/"
print(f"Processing all PDF reference files in directory {target_dir}")

stdout = check_output(f"git diff --name-status {GIT_REF}", shell=True)
changed_pdf_files = [
    line[1:].strip()
    for line in stdout.decode("utf-8").splitlines()
    if line.startswith(f"M\t{target_dir}") and line.endswith(".pdf")
]
changes_pdf_files_count = len(changed_pdf_files)
changed_pdf_files = changed_pdf_files[VIEW_SLICE]
is_shrunk = len(changed_pdf_files) < changes_pdf_files_count

for changed_pdf_file in changed_pdf_files:
    (TMP_DIR / Path(changed_pdf_file)).parent.mkdir(exist_ok=True, parents=True)
    command = f"git show {GIT_REF}:{changed_pdf_file} > {TMP_DIR}/{changed_pdf_file}"
    print(command)
    check_output(command, shell=True)

env = Environment(
    loader=FileSystemLoader(str(SCRIPTS_DIR)),
    autoescape=True,
    trim_blocks=True,
    lstrip_blocks=True,
)
template = env.get_template(TEMPLATE_FILENAME)
(REPO_DIR / TEMPLATE_FILENAME).write_text(
    template.render(
        changed_pdf_files=changed_pdf_files,
        is_shrunk=is_shrunk,
        changes_pdf_files_count=changes_pdf_files_count,
        GIT_REF=GIT_REF,
        VIEW_SLICE=VIEW_SLICE,
    )
)

httpd = HTTPServer(
    ("", PORT), partial(SimpleHTTPRequestHandler, directory=str(REPO_DIR))
)
webbrowser.open(f"http://localhost:{PORT}/{TEMPLATE_FILENAME}")
httpd.serve_forever()



"""
Routines for organizing lines and larger blocks of text, with manual and
automatic line wrapping.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/LineBreaks.html>
"""

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import uuid4

from fpdf.drawing_primitives import DeviceCMYK, DeviceGray, DeviceRGB

from .enums import Align, CharVPos, TextDirection, TextMode, WrapMode
from .errors import FPDFException
from .fonts import CoreFont, TTFFont
from .graphics_state import GraphicsState
from .util import FloatTolerance, escape_parens

StateStackType = GraphicsState

SOFT_HYPHEN = "\u00ad"
HYPHEN = "\u002d"
SPACE = " "
BREAKING_SPACE_SYMBOLS = [
    " ",
    "\u200b",  # | ZERO WIDTH SPACE
    "\u2000",  # | EN QUAD
    "\u2001",  # | EM QUAD
    "\u2002",  # | EN SPACE
    "\u2003",  # | EM SPACE
    "\u2004",  # | THREE-PER-EM SPACE
    "\u2005",  # | FOUR-PER-EM SPACE
    "\u2006",  # | SIX-PER-EM SPACE
    "\u2008",  # | PUNCTUATION SPACE
    "\u2009",  # | THIN SPACE
    "\u200a",  # | HAIR SPACE
    "\u205f",  # | MEDIUM MATHEMATICAL SPACE
    "\u3000",  # | IDEOGRAPHIC SPACE
    "\u0009",  # | TAB
]
BREAKING_SPACE_SYMBOLS_STR = "".join(BREAKING_SPACE_SYMBOLS)
NBSP = "\u00a0"
NEWLINE = "\n"
FORM_FEED = "\u000c"


class Fragment:
    """
    A fragment of text with font/size/style and other associated information.
    """

    def __init__(
        self,
        characters: Union[list[str], str],
        graphics_state: StateStackType,
        k: float,
        link: Optional[int | str] = None,
    ) -> None:
        if isinstance(characters, str):
            self.characters = list(characters)
        else:
            self.characters = characters
        self.graphics_state = graphics_state
        self.k = k
        self.link = link

    def __repr__(self) -> str:
        return (
            f"Fragment(characters={self.characters},"
            f" graphics_state={self.graphics_state},"
            f" k={self.k}, link={self.link})"
        )

    @property
    def font(self) -> CoreFont | TTFFont:
        if TYPE_CHECKING:
            assert self.graphics_state.current_font is not None
        return self.graphics_state.current_font

    @font.setter
    def font(self, v: CoreFont | TTFFont) -> None:
        self.graphics_state.current_font = v

    @property
    def is_ttf_font(self) -> bool:
        return self.font is not None and self.font.type == "TTF"

    @property
    def font_style(self) -> str:
        return self.graphics_state.font_style

    @property
    def font_family(self) -> str:
        return self.graphics_state.font_family

    @property
    def font_size_pt(self) -> float:
        size = self.graphics_state.font_size_pt
        vpos = self.graphics_state.char_vpos
        if vpos == CharVPos.SUB:
            size *= self.graphics_state.sub_scale
        elif vpos == CharVPos.SUP:
            size *= self.graphics_state.sup_scale
        elif vpos == CharVPos.NOM:
            size *= self.graphics_state.nom_scale
        elif vpos == CharVPos.DENOM:
            size *= self.graphics_state.denom_scale
        return size

    @property
    def font_size(self) -> float:
        return self.graphics_state.font_size_pt / self.k

    @property
    def font_stretching(self) -> float:
        return self.graphics_state.font_stretching

    @property
    def char_spacing(self) -> float:
        return self.graphics_state.char_spacing

    @property
    def text_mode(self) -> TextMode:
        return self.graphics_state.text_mode

    @property
    def underline(self) -> bool:
        return self.graphics_state.underline

    @property
    def strikethrough(self) -> bool:
        return self.graphics_state.strikethrough

    @property
    def draw_color(self) -> Optional[DeviceRGB | DeviceGray | DeviceCMYK]:
        return self.graphics_state.draw_color

    @property
    def fill_color(self) -> Optional[DeviceRGB | DeviceGray | DeviceCMYK]:
        return self.graphics_state.fill_color

    @property
    def text_color(self) -> Optional[DeviceRGB | DeviceGray | DeviceCMYK]:
        return self.graphics_state.text_color

    @property
    def line_width(self) -> float:
        return self.graphics_state.line_width

    @property
    def char_vpos(self) -> CharVPos:
        return self.graphics_state.char_vpos

    @property
    def lift(self) -> float:
        vpos = self.graphics_state.char_vpos
        if vpos == CharVPos.SUB:
            lift: float = self.graphics_state.sub_lift
        elif vpos == CharVPos.SUP:
            lift = self.graphics_state.sup_lift
        elif vpos == CharVPos.NOM:
            lift = self.graphics_state.nom_lift
        elif vpos == CharVPos.DENOM:
            lift = self.graphics_state.denom_lift
        else:
            lift = 0.0
        return lift * self.graphics_state.font_size_pt

    @property
    def string(self) -> str:
        return "".join(self.characters)

    @property
    def width(self) -> float:
        return self.get_width()

    @property
    def text_shaping_parameters(self) -> Optional[Dict[str, Any]]:
        return self.graphics_state.text_shaping

    @property
    def paragraph_direction(self) -> TextDirection:
        if TYPE_CHECKING:
            assert self.text_shaping_parameters is not None
            assert isinstance(
                self.text_shaping_parameters["paragraph_direction"], TextDirection
            )
        return (
            self.text_shaping_parameters["paragraph_direction"]
            if self.text_shaping_parameters
            else TextDirection.LTR
        )

    @property
    def fragment_direction(self) -> TextDirection:
        if TYPE_CHECKING:
            assert self.text_shaping_parameters is not None
            assert isinstance(
                self.text_shaping_parameters["fragment_direction"], TextDirection
            )
        return (
            self.text_shaping_parameters["fragment_direction"]
            if self.text_shaping_parameters
            else TextDirection.LTR
        )

    def trim(self, index: int) -> None:
        self.characters = self.characters[:index]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Fragment):
            return False
        return (
            self.characters == other.characters
            and self.graphics_state == other.graphics_state
            and self.k == other.k
        )

    def __hash__(self) -> int:
        return hash((self.characters, self.graphics_state, self.k))

    def get_width(
        self,
        start: int = 0,
        end: Optional[int] = None,
        chars: Optional[str] = None,
        initial_cs: bool = True,
    ) -> float:
        """
        Return the width of the string with the given font/size/style/etc.

        Args:
            start (int): Index of the start character. Default start of fragment.
            end (int): Index of the end character. Default end of fragment.
            chars (str): Specific text to get the width for (not necessarily the
                same as the contents of the fragment). If given, this takes
                precedence over the start/end arguments.
        """

        if chars is None:
            if end is not None:
                chars = "".join(self.characters[start:end])
            else:
                chars = "".join(self.characters[start:])
        char_len, w = self.font.get_text_width(
            chars, self.font_size_pt, self.text_shaping_parameters
        )
        char_spacing = self.char_spacing
        if self.font_stretching != 100:
            w *= self.font_stretching * 0.01
            char_spacing *= self.font_stretching * 0.01
        if self.char_spacing != 0:
            # initial_cs must be False if the fragment is located at the
            # beginning of a text object, because the first char won't get spaced.
            if initial_cs:
                w += char_spacing * char_len
            else:
                w += char_spacing * (char_len - 1)
        return w / self.k

    def has_same_style(self, other: "Fragment") -> bool:
        """Returns if 2 fragments are equivalent other than the characters/string"""
        return (
            self.graphics_state == other.graphics_state
            and self.k == other.k
            and self.__class__ == other.__class__
        )

    def get_character_width(
        self, character: str, print_sh: bool = False, initial_cs: bool = True
    ) -> float:
        """
        Return the width of a single character out of the stored text.
        """
        if character == SOFT_HYPHEN and not print_sh:
            # HYPHEN is inserted instead of SOFT_HYPHEN
            character = HYPHEN
        return self.get_width(chars=character, initial_cs=initial_cs)

    def render_pdf_text(
        self,
        frag_ws: float,
        current_ws: float,
        word_spacing: float,
        adjust_x: float,
        adjust_y: float,
        h: float,
    ) -> str:
        if self.is_ttf_font:
            if self.text_shaping_parameters:
                return self.render_with_text_shaping(
                    adjust_x, adjust_y, h, word_spacing
                )
            return self.render_pdf_text_ttf(frag_ws, word_spacing)
        return self.render_pdf_text_core(frag_ws, current_ws)

    def render_pdf_text_ttf(self, frag_ws: float, word_spacing: float) -> str:
        assert isinstance(self.font, TTFFont)
        ret = ""
        mapped_text = ""
        for char in self.string:
            mapped_char = self.font.subset.pick(ord(char))
            if mapped_char:
                mapped_text += chr(mapped_char)
        if word_spacing:
            # do this once in advance
            u_space = self.font.escape_text(" ")

            # According to the PDF reference, word spacing shall be applied to every
            # occurrence of the single-byte character code 32 in a string when using
            # a simple font or a composite font that defines code 32 as a single-byte code.
            # It shall not apply to occurrences of the byte value 32 in multiple-byte codes.
            # FPDF uses 2 bytes per character (UTF-16-BE encoding) so the "Tw" operator doesn't work
            # As a workaround, we do word spacing using an adjustment before each space.
            # Determine the index of the space character (" ") in the current
            # subset and split words whenever this mapping code is found
            #
            space_char_id = self.font.subset.pick(ord(" "))
            assert space_char_id is not None
            words = mapped_text.split(chr(space_char_id))
            words_strl: list[str] = []
            for word_i, word in enumerate(words):
                # pylint: disable=redefined-loop-name
                word = self.font.escape_text(word)
                if word_i == 0:
                    words_strl.append(f"({word})")
                else:
                    adj = -(frag_ws * self.k) * 1000 / self.font_size_pt
                    words_strl.append(f"{adj:.3f}({u_space}{word})")
            escaped_text = " ".join(words_strl)
            ret += f"[{escaped_text}] TJ"
        else:
            escaped_text = self.font.escape_text(mapped_text)
            ret += f"({escaped_text}) Tj"
        return ret

    def render_with_text_shaping(
        self, pos_x: float, pos_y: float, h: float, word_spacing: float
    ) -> str:
        assert isinstance(self.font, TTFFont)
        ret = ""
        text = ""
        space_mapped_code = self.font.subset.pick(ord(" "))

        def adjust_pos(pos: float) -> float:
            if TYPE_CHECKING:
                assert isinstance(self.font, TTFFont)
            return (
                pos
                * self.font.scale
                * self.font_size_pt
                * (self.font_stretching / 100)
                / 1000
                / self.k
            )

        char_spacing = self.char_spacing * (self.font_stretching / 100) / self.k
        for ti in self.font.shape_text(
            self.string, self.font_size_pt, self.text_shaping_parameters
        ):
            if ti["mapped_char"] is None:  # Missing glyph
                continue
            char = self.font.escape_text(chr(ti["mapped_char"]))
            if ti["x_offset"] != 0 or ti["y_offset"] != 0:
                if text:
                    ret += f"({text}) Tj "
                    text = ""
                offsetx = pos_x + adjust_pos(ti["x_offset"])
                offsety = pos_y - adjust_pos(ti["y_offset"])
                ret += (
                    f"1 0 0 1 {(offsetx) * self.k:.2f} {(h - offsety) * self.k:.2f} Tm "
                )
            text += char
            pos_x += adjust_pos(ti["x_advance"]) + char_spacing
            pos_y += adjust_pos(ti["y_advance"])
            if word_spacing and ti["mapped_char"] == space_mapped_code:
                pos_x += word_spacing

            # if only moving "x" we don't need to move the text matrix
            if ti["force_positioning"] or (
                word_spacing and ti["mapped_char"] == space_mapped_code
            ):
                if text:
                    ret += f"({text}) Tj "
                    text = ""
                ret += f"1 0 0 1 {(pos_x) * self.k:.2f} {(h - pos_y) * self.k:.2f} Tm "

        if text:
            ret += f"({text}) Tj"
        return ret

    def render_pdf_text_core(self, frag_ws: float, current_ws: float) -> str:
        ret = ""
        if frag_ws != current_ws:
            ret += f"{frag_ws * self.k:.3f} Tw "
        escaped_text = escape_parens(self.string)
        ret += f"({escaped_text}) Tj"
        return ret


class TotalPagesSubstitutionFragment(Fragment):
    """
    A special type of text fragment that represents a placeholder for the total number of pages
    in a PDF document.

    A placeholder will be generated during the initial content rendering phase of a PDF document.
    This placeholder is later replaced by the total number of pages in the document when the final
    output is being produced.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.uuid = uuid4()

    def get_placeholder_string(self) -> str:
        """
        This method returns a placeholder string containing a universally unique identifier (UUID4),
        ensuring that the placeholder is distinct and does not conflict with other placeholders
        within the document.
        """
        return f"::placeholder:{self.uuid}::"

    def render_pdf_text(self, *args: Any, **kwargs: Any) -> str:
        """
        This method is invoked during the page content rendering phase, which is common to all
        `Fragment` instances. It stores the provided arguments and keyword arguments to preserve
        the necessary information and graphic state for the final substitution rendering.

        The method then returns the unique placeholder string.
        """
        self._render_args = args
        self._render_kwargs = kwargs
        return self.get_placeholder_string()

    def render_text_substitution(self, replacement_text: str) -> str:
        """
        This method is invoked at the output phase. It calls `render_pdf_text()` from the superclass
        to render the fragment with the preserved rendering state (stored in `_render_args` and `_render_kwargs`)
        and insert the final text in place of the placeholder.
        """
        self.characters = list(replacement_text)
        return super().render_pdf_text(*self._render_args, **self._render_kwargs)


class TextLine(NamedTuple):
    fragments: Sequence[Fragment]
    text_width: float
    number_of_spaces: int
    align: Align
    height: float
    max_width: Optional[float]
    trailing_nl: bool = False
    trailing_form_feed: bool = False
    indent: float = 0

    def get_ordered_fragments(self) -> List[Fragment]:
        if not self.fragments:
            return []
        directional_runs: list[list[Fragment]] = []
        direction = None
        for fragment in self.fragments:
            if direction is not None and fragment.fragment_direction == direction:
                directional_runs[-1].append(fragment)
            else:
                directional_runs.append([fragment])
                direction = fragment.fragment_direction
        if self.fragments[0].paragraph_direction == TextDirection.RTL or (
            not self.fragments[0].paragraph_direction
            and self.fragments[0].fragment_direction == TextDirection.RTL
        ):
            directional_runs = directional_runs[::-1]
        ordered_fragments: list[Fragment] = []
        for run in directional_runs:
            ordered_fragments += (
                run[::-1] if run[0].fragment_direction == TextDirection.RTL else run
            )
        return ordered_fragments


class SpaceHint(NamedTuple):
    original_fragment_index: int
    original_character_index: int
    current_line_fragment_index: int
    current_line_character_index: int
    line_width: float
    number_of_spaces: int


class HyphenHint(NamedTuple):
    original_fragment_index: int
    original_character_index: int
    current_line_fragment_index: int
    current_line_character_index: int
    line_width: float
    number_of_spaces: int
    curchar: str
    curchar_width: float
    graphics_state: StateStackType
    k: float


class CurrentLine:
    def __init__(
        self, max_width: float, print_sh: bool = False, indent: float = 0
    ) -> None:
        """
        Per-line text fragment management for use by MultiLineBreak.
            Args:
                print_sh (bool): If true, a soft-hyphen will be rendered
                    normally, instead of triggering a line break. Default: False
        """
        self.max_width = max_width
        self.print_sh = print_sh
        self.indent = indent
        self.fragments: List[Fragment] = []
        self.height: float = 0
        self.number_of_spaces: int = 0

        # automatic break hints
        # CurrentLine class remembers 3 positions
        # 1 - position of last inserted character.
        #     class attributes (`width`, `fragments`)
        #     is used for this purpose
        # 2 - position of last inserted space
        #     SpaceHint is used for this purpose.
        # 3 - position of last inserted soft-hyphen
        #     HyphenHint is used for this purpose.
        # The purpose of multiple positions tracking - to have an ability
        # to break in multiple places, depending on condition.
        self.space_break_hint: Optional[SpaceHint] = None
        self.hyphen_break_hint: Optional[HyphenHint] = None

    @property
    def width(self) -> float:
        width: float = 0
        for i, fragment in enumerate(self.fragments):
            width += fragment.get_width(initial_cs=i > 0)
        return width

    def add_character(
        self,
        character: str,
        character_width: float,
        original_fragment: Fragment | HyphenHint,
        original_fragment_index: int,
        original_character_index: int,
        height: float,
        url: Optional[str | int] = None,
    ) -> None:
        assert character != NEWLINE
        self.height = height
        if not self.fragments:
            assert isinstance(original_fragment, Fragment)
            self.fragments.append(
                original_fragment.__class__(
                    characters="",
                    graphics_state=original_fragment.graphics_state,
                    k=original_fragment.k,
                    link=url,
                )
            )

        # characters are expected to be grouped into fragments by font and
        # character attributes. If the last existing fragment doesn't match
        # the properties of the pending character -> add a new fragment.
        elif isinstance(original_fragment, Fragment):
            if isinstance(
                self.fragments[-1], Fragment
            ) and not original_fragment.has_same_style(self.fragments[-1]):
                self.fragments.append(
                    original_fragment.__class__(
                        characters="",
                        graphics_state=original_fragment.graphics_state,
                        k=original_fragment.k,
                        link=url,
                    )
                )
        active_fragment = self.fragments[-1]

        if character in BREAKING_SPACE_SYMBOLS_STR:
            self.space_break_hint = SpaceHint(
                original_fragment_index,
                original_character_index,
                len(self.fragments),
                len(active_fragment.characters),
                self.width,
                self.number_of_spaces,
            )
            self.number_of_spaces += 1
        elif character == NBSP:
            # PDF viewers ignore NBSP for word spacing with "Tw".
            character = SPACE
            self.number_of_spaces += 1
        elif character == SOFT_HYPHEN and not self.print_sh:
            self.hyphen_break_hint = HyphenHint(
                original_fragment_index,
                original_character_index,
                len(self.fragments),
                len(active_fragment.characters),
                self.width,
                self.number_of_spaces,
                HYPHEN,
                character_width,
                original_fragment.graphics_state,
                original_fragment.k,
            )

        if character != SOFT_HYPHEN or self.print_sh:
            active_fragment.characters.append(character)

    def trim_trailing_spaces(self) -> None:
        if not self.fragments:
            return
        last_frag = self.fragments[-1]
        last_char = last_frag.characters[-1]
        while last_char == " ":
            last_frag.trim(-1)
            if not last_frag.characters:
                del self.fragments[-1]
            if not self.fragments:
                return
            last_frag = self.fragments[-1]
            last_char = last_frag.characters[-1]

    def _apply_automatic_hint(self, break_hint: SpaceHint | HyphenHint) -> None:
        """
        This function mutates the current_line, applying one of the states
        observed in the past and stored in
        `hyphen_break_hint` or `space_break_hint` attributes.
        """
        self.fragments = self.fragments[: break_hint.current_line_fragment_index]
        if self.fragments:
            self.fragments[-1].trim(break_hint.current_line_character_index)
        self.number_of_spaces = break_hint.number_of_spaces

    def manual_break(
        self, align: Align, trailing_nl: bool = False, trailing_form_feed: bool = False
    ) -> TextLine:
        return TextLine(
            fragments=self.fragments,
            text_width=self.width,
            number_of_spaces=self.number_of_spaces,
            align=align,
            height=self.height,
            max_width=self.max_width - self.indent,
            trailing_nl=trailing_nl,
            trailing_form_feed=trailing_form_feed,
            indent=self.indent,
        )

    def automatic_break_possible(self) -> bool:
        return self.hyphen_break_hint is not None or self.space_break_hint is not None

    def automatic_break(self, align: Align) -> Tuple[int, int, TextLine]:
        assert self.automatic_break_possible()
        if self.hyphen_break_hint is not None and (
            self.space_break_hint is None
            or self.hyphen_break_hint.line_width > self.space_break_hint.line_width
        ):
            self._apply_automatic_hint(self.hyphen_break_hint)
            self.add_character(
                self.hyphen_break_hint.curchar,
                self.hyphen_break_hint.curchar_width,
                self.hyphen_break_hint,
                self.hyphen_break_hint.original_fragment_index,
                self.hyphen_break_hint.original_character_index,
                self.height,
            )
            return (
                self.hyphen_break_hint.original_fragment_index,
                self.hyphen_break_hint.original_character_index,
                self.manual_break(align),
            )
        assert self.space_break_hint is not None
        self._apply_automatic_hint(self.space_break_hint)
        return (
            self.space_break_hint.original_fragment_index,
            self.space_break_hint.original_character_index,
            self.manual_break(align),
        )


class MultiLineBreak:
    def __init__(
        self,
        fragments: Sequence[Fragment],
        max_width: Union[float, Callable[[float], float]],
        margins: Sequence[float],
        align: Align = Align.L,
        print_sh: bool = False,
        wrapmode: WrapMode = WrapMode.WORD,
        line_height: float = 1.0,
        skip_leading_spaces: bool = False,
        first_line_indent: float = 0,
    ):
        """Accept text as Fragments, to be split into individual lines depending
        on line width and text height.
        Args:
            fragments: A sequence of Fragment()s containing text.
            max_width: Either a fixed width as float or a callback function
                get_width(height). If a function, it gets called with the largest
                height encountered on the current line, and must return the
                applicable width for the line with the given height at the current
                vertical position. The height is relevant in those cases where the
                lateral boundaries of the enclosing TextRegion() are not vertical.
            margins (sequence of floats): The extra clearance that may apply at the beginning
                and/or end of a line (usually either FPDF.c_margin or 0.0 for each side).
            align (Align): The horizontal alignment of the current text block.
            print_sh (bool): If True, a soft-hyphen will be rendered
                normally, instead of triggering a line break. Default: False
            wrapmode (WrapMode): Selects word or character based wrapping.
            line_height (float, optional): A multiplier relative to the font
                size changing the vertical space occupied by a line of text. Default 1.0.
            skip_leading_spaces (bool, optional): On each line, any space characters
                at the beginning will be skipped. Default value: False.
            first_line_indent (float, optional): left spacing before first line of text in paragraph.
        """
        self.get_width: Callable[[float], float]
        self.fragments = fragments
        if callable(max_width):
            self.get_width = max_width
        else:
            self.get_width = lambda height: max_width
        self.margins = margins
        self.align = align
        self.print_sh = print_sh
        self.wrapmode = wrapmode
        self.line_height = line_height
        self.skip_leading_spaces = skip_leading_spaces
        self.fragment_index: int = 0
        self.character_index: int = 0
        self.idx_last_forced_break: Optional[int] = None
        self.first_line_indent = first_line_indent
        self._is_first_line = True

    # pylint: disable=too-many-return-statements
    def get_line(self) -> Optional[TextLine]:
        first_char = True  # "Tw" ignores the first character in a text object.
        idx_last_forced_break = self.idx_last_forced_break
        self.idx_last_forced_break = None

        if self.fragment_index == len(self.fragments):
            return None

        current_font_height: float = 0

        max_width = self.get_width(current_font_height)
        # The full max width will be passed on via TextLine to FPDF._render_styled_text_line().
        current_line = CurrentLine(
            max_width=max_width,
            print_sh=self.print_sh,
            indent=self.first_line_indent if self._is_first_line else 0,
        )
        # For line wrapping we need to use the reduced width.
        for margin in self.margins:
            max_width -= float(margin)
        if self._is_first_line:
            max_width -= self.first_line_indent

        if self.skip_leading_spaces:
            # write_html() with TextColumns uses this, since it can't know in
            # advance where the lines will be broken.
            while self.fragment_index < len(self.fragments):
                if self.character_index >= len(
                    self.fragments[self.fragment_index].characters
                ):
                    self.character_index = 0
                    self.fragment_index += 1
                    continue
                character = self.fragments[self.fragment_index].characters[
                    self.character_index
                ]
                if character == SPACE:
                    self.character_index += 1
                else:
                    break

        while self.fragment_index < len(self.fragments):
            current_fragment = self.fragments[self.fragment_index]

            if FloatTolerance.greater_than(
                current_fragment.font_size, current_font_height
            ):
                current_font_height = current_fragment.font_size  # document units
                max_width = self.get_width(current_font_height)
                current_line.max_width = max_width
                for margin in self.margins:
                    max_width -= float(margin)
                if self._is_first_line:
                    max_width -= self.first_line_indent

            if self.character_index >= len(current_fragment.characters):
                self.character_index = 0
                self.fragment_index += 1
                continue

            character = current_fragment.characters[self.character_index]
            character_width = current_fragment.get_character_width(
                character, self.print_sh, initial_cs=not first_char
            )
            first_char = False

            if character in (NEWLINE, FORM_FEED):
                self.character_index += 1
                if not current_line.fragments:
                    current_line.height = current_font_height * self.line_height
                self._is_first_line = False
                return current_line.manual_break(
                    Align.L if self.align == Align.J else self.align,
                    trailing_nl=character == NEWLINE,
                    trailing_form_feed=character == FORM_FEED,
                )
            if FloatTolerance.greater_than(
                current_line.width + character_width, max_width
            ):
                self._is_first_line = False
                if (
                    character in BREAKING_SPACE_SYMBOLS_STR
                ):  # must come first, always drop a current space.
                    self.character_index += 1
                    return current_line.manual_break(self.align)
                if self.wrapmode == WrapMode.CHAR:
                    # If the line ends with one or more spaces, then we want to get
                    # rid of them so it can be justified correctly.
                    current_line.trim_trailing_spaces()
                    return current_line.manual_break(self.align)
                if current_line.automatic_break_possible():
                    (
                        self.fragment_index,
                        self.character_index,
                        line,
                    ) = current_line.automatic_break(self.align)
                    self.character_index += 1
                    return line
                if idx_last_forced_break == self.character_index:
                    raise FPDFException(
                        "Not enough horizontal space to render a single character"
                    )
                self.idx_last_forced_break = self.character_index
                return current_line.manual_break(
                    Align.L if self.align == Align.J else self.align,
                )

            current_line.add_character(
                character,
                character_width,
                current_fragment,
                self.fragment_index,
                self.character_index,
                current_font_height * self.line_height,
                current_fragment.link,
            )

            self.character_index += 1

        if current_line.width:
            self._is_first_line = False
            return current_line.manual_break(
                Align.L if self.align == Align.J else self.align,
            )
        return None



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


IMG_FILE_PATH = HERE / "../image_types/insert_images_insert_png.png"
IMG_DESCRIPTION = "Democratic Socialists of America Logo"


def test_alt_text_and_title(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(IMG_FILE_PATH, alt_text=IMG_DESCRIPTION)
    pdf.image(IMG_FILE_PATH, title=IMG_DESCRIPTION)
    pdf.image(IMG_FILE_PATH, alt_text=IMG_DESCRIPTION, title=IMG_DESCRIPTION)
    assert_pdf_equal(pdf, HERE / "alt_text_and_title.pdf", tmp_path)


def test_alt_text_on_two_pages(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(IMG_FILE_PATH, alt_text=IMG_DESCRIPTION)
    pdf.add_page()
    pdf.image(IMG_FILE_PATH, alt_text=IMG_DESCRIPTION)
    assert_pdf_equal(pdf, HERE / "test_alt_text_on_two_pages.pdf", tmp_path)



import warnings
from abc import ABC
from typing import TYPE_CHECKING, Optional, Union

from .syntax import PDFString, build_obj_dict, create_dictionary_string

if TYPE_CHECKING:
    from .encryption import StandardSecurityHandler
    from .syntax import Destination


class Action(ABC):
    def __init__(self, next_action: Optional["Action"] = None) -> None:
        """
        Args:
            next (PDFObject | str): optional reference to another Action to trigger after this one
        """
        self.next = next_action

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        raise NotImplementedError

    def _serialize(
        self,
        key_values: Optional[dict[str, object]] = None,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        if key_values is None:
            key_values = {}
        if self.next:
            key_values["Next"] = self.next
        obj_dict = build_obj_dict(key_values, _security_handler, _obj_id)
        return create_dictionary_string(obj_dict, field_join=" ")


class URIAction(Action):
    def __init__(self, uri: str, next_action: Optional[Action] = None) -> None:
        super().__init__(next_action)
        self.uri = uri

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return super()._serialize(
            {"s": "/URI", "u_r_i": PDFString(self.uri, encrypt=True)},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )


class NamedAction(Action):
    def __init__(self, action_name: str, next_action: Optional[Action] = None) -> None:
        super().__init__(next_action)
        if action_name not in ("NextPage", "PrevPage", "FirstPage", "LastPage"):
            warnings.warn("Non-standard named action added")
        self.action_name = action_name

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return super()._serialize(
            {"s": "/Named", "n": f"/{self.action_name}"},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )


class GoToAction(Action):
    def __init__(
        self, dest: Union[str, "Destination"], next_action: Optional[Action] = None
    ) -> None:
        super().__init__(next_action)
        if isinstance(dest, str) and dest.startswith("#"):
            dest = PDFString(dest[1:], encrypt=True)
        self.dest = dest

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return super()._serialize(
            {"s": "/GoTo", "d": self.dest},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )


class GoToRemoteAction(Action):
    def __init__(
        self, file: str, dest: "Destination", next_action: Optional[Action] = None
    ) -> None:
        super().__init__(next_action)
        self.file = file
        self.dest = dest

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return super()._serialize(
            {"s": "/GoToR", "f": PDFString(self.file, encrypt=True), "d": self.dest},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )


class LaunchAction(Action):
    "As of 2022, this does not seem honored by neither Adobe Acrobat nor Sumatra readers."

    def __init__(self, file: str, next_action: Optional[Action] = None) -> None:
        super().__init__(next_action)
        self.file = file

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        return super()._serialize(
            {"s": "/Launch", "f": PDFString(self.file, encrypt=True)},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )


# Annotation & actions that we tested implementing,
# but that revealed not be worth the effort:
# * Popup annotation & Hide action: as of june 2021,
#   do not seem support neither by Adobe Acrobat nor by Sumatra.
#   Moreover, they both use to indirect reference annotations,
#   and hence implementing them would need some consequent refactoring,
#   as annotations are currently defined "inline", not as dedicated PDF objects.



"""
Various utilities that could not be gathered logically in a specific module.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.
"""

import decimal

# nosemgrep: python.lang.compatibility.python37.python37-compatibility-importlib2 (min Python is 3.9)
from importlib import resources
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Iterable,
    NamedTuple,
    Sequence,
    TypeVar,
    Union,
    overload,
)

if TYPE_CHECKING:
    from PIL.Image import Image as PILImage

    from .svg import SVGObject

ImageType = Union[str, bytes, BinaryIO, "PILImage", Path, None]
ImageClass = (str, bytes, BinaryIO, "PILImage", Path)
ImageData = Union["SVGObject", "PILImage", bytes, BinaryIO, Path, None]
SVGObjectType = TypeVar("SVGObjectType", bound="SVGObject")
Number = Union[int, float, decimal.Decimal]
NumberClass = (int, float, decimal.Decimal)
_StrBytes = TypeVar("_StrBytes", str, bytes)


class Padding(NamedTuple):
    top: float = 0
    right: float = 0
    bottom: float = 0
    left: float = 0

    @classmethod
    def new(cls, padding: Union[Number, Sequence[Number], "Padding"]) -> "Padding":
        """Return a 4-tuple of padding values from a single value or a 2, 3 or 4-tuple according to CSS rules"""
        if isinstance(padding, NumberClass):
            return Padding(
                float(padding), float(padding), float(padding), float(padding)
            )
        if len(padding) == 2:
            return Padding(
                float(padding[0]),
                float(padding[1]),
                float(padding[0]),
                float(padding[1]),
            )
        if len(padding) == 3:
            return Padding(
                float(padding[0]),
                float(padding[1]),
                float(padding[2]),
                float(padding[1]),
            )
        if len(padding) == 4:
            return Padding(
                float(padding[0]),
                float(padding[1]),
                float(padding[2]),
                float(padding[3]),
            )

        raise ValueError(
            f"padding shall be a number or a sequence of 2, 3 or 4 numbers, got {str(padding)}"
        )


def buffer_subst(buffer: bytearray, placeholder: str, value: str) -> bytearray:
    buffer_size = len(buffer)
    assert len(placeholder) == len(value), f"placeholder={placeholder} value={value}"
    buffer = buffer.replace(placeholder.encode(), value.encode(), 1)
    assert len(buffer) == buffer_size
    return buffer


@overload
def escape_parens(s: str) -> str: ...


@overload
def escape_parens(s: bytes) -> bytes: ...


def escape_parens(s: _StrBytes) -> _StrBytes:
    """Add a backslash character before , ( and )"""
    if isinstance(s, str):
        return (
            s.replace("\\", "\\\\")
            .replace(")", "\\)")
            .replace("(", "\\(")
            .replace("\r", "\\r")
        )
    return (
        s.replace(b"\\", b"\\\\")
        .replace(b")", b"\\)")
        .replace(b"(", b"\\(")
        .replace(b"\r", b"\\r")
    )


def get_scale_factor(unit: Union[str, Number]) -> float:
    """
    Get how many pts are in a unit. (k)

    Args:
        unit (str, float, int): Any of "pt", "mm", "cm", "in", or a number.
    Returns:
        float: The number of points in that unit (assuming 72dpi)
    Raises:
        ValueError
    """
    if isinstance(unit, NumberClass):
        return float(unit)

    if unit == "pt":
        return 1
    if unit == "mm":
        return 72 / 25.4
    if unit == "cm":
        return 72 / 2.54
    if unit == "in":
        return 72.0
    raise ValueError(f"Incorrect unit: {unit}")


def convert_unit(
    to_convert: Number | Iterable[Any],
    old_unit: Union[str, Number],
    new_unit: Union[str, Number],
) -> Union[float, tuple[Any, ...]]:
    """
     Convert a number or sequence of numbers from one unit to another.

     If either unit is a number it will be treated as the number of points per unit.  So 72 would mean 1 inch.

     Args:
        to_convert (float, int, Iterable): The number / list of numbers, or points, to convert
        old_unit (str, float, int): A unit accepted by `fpdf.fpdf.FPDF` or a number
        new_unit (str, float, int): A unit accepted by `fpdf.fpdf.FPDF` or a number
    Returns:
        (float, tuple): to_convert converted from old_unit to new_unit or a tuple of the same
    """
    unit_conversion_factor = get_scale_factor(new_unit) / get_scale_factor(old_unit)
    if isinstance(to_convert, Iterable):
        return tuple(convert_unit(i, 1, unit_conversion_factor) for i in to_convert)
    return float(to_convert) / unit_conversion_factor


def number_to_str(number: Number) -> str:
    """
    Convert a decimal number to a minimal string representation (no trailing 0 or .).

    Args:
        number (Number): the number to be converted to a string.

    Returns:
        The number's string representation.
    """
    # this approach tries to produce minimal representations of floating point numbers
    # but can also produce "-0".
    return f"{number:.4f}".rstrip("0").rstrip(".")


ROMAN_NUMERAL_MAP = (
    ("M", 1000),
    ("CM", 900),
    ("D", 500),
    ("CD", 400),
    ("C", 100),
    ("XC", 90),
    ("L", 50),
    ("XL", 40),
    ("X", 10),
    ("IX", 9),
    ("V", 5),
    ("IV", 4),
    ("I", 1),
)


def int2roman(n: int) -> str:
    "Convert an integer to Roman numeral"
    result = ""
    if n is None:
        return result
    for numeral, integer in ROMAN_NUMERAL_MAP:
        while n >= integer:
            result += numeral
            n -= integer
    return result


def int_to_letters(n: int) -> str:
    "Convert an integer to a letter value (A to Z for the first 26, then AA to ZZ, and so on)"
    if n > 25:
        return int_to_letters(int((n / 26) - 1)) + int_to_letters(n % 26)
    return chr(n + ord("A"))


def builtin_srgb2014_bytes() -> bytes:
    pkg = "fpdf.data.color_profiles"
    return (resources.files(pkg) / "sRGB2014.icc").read_bytes()


def format_number(x: float, digits: int = 8) -> str:
    # snap tiny values to zero to avoid "-0" and scientific notation
    if abs(x) < 1e-12:
        x = 0.0
    s = f"{x:.{digits}f}"
    if "." in s:
        s = s.rstrip("0").rstrip(".")
    if s == "-0":
        s = "0"
    if s.startswith("."):
        s = "0" + s
    if s.startswith("-."):
        s = s.replace("-.", "-0.", 1)
    return s


def get_parsed_unicode_range(
    unicode_range: str | Sequence[str | int | tuple[int, int]],
) -> set[int]:
    """
    Parse unicode_range parameter into a set of codepoints.

    Supports CSS-style formats:

    - String with comma-separated ranges: "U+1F600-1F64F, U+2600-26FF, U+2615"
    - List of strings: ["U+1F600-1F64F", "U+2600", "U+26FF"]
    - List of tuples: [(0x1F600, 0x1F64F), (0x2600, 0x26FF)]
    - List of integers: [0x1F600, 0x2600, 128512]
    - Mixed formats: [(0x1F600, 0x1F64F), "U+2600", 128512]

    Returns a set of integer codepoints.
    """
    if unicode_range is not None and len(unicode_range) == 0:
        raise ValueError("unicode_range cannot be empty")

    codepoints: set[int] = set()

    if isinstance(unicode_range, str):
        unicode_range = [item.strip() for item in unicode_range.split(",")]

    for item in unicode_range:
        if isinstance(item, tuple):
            if len(item) != 2:
                raise ValueError(f"Tuple must have exactly 2 elements: {item}")
            start, end = item

            if isinstance(start, str):
                start = int(start.replace("U+", "").replace("u+", ""), 16)
            if isinstance(end, str):
                end = int(end.replace("U+", "").replace("u+", ""), 16)

            if start > end:
                raise ValueError(f"Invalid range: start ({start}) > end ({end})")

            codepoints.update(range(start, end + 1))

        elif isinstance(item, str):
            item_stripped = item.strip().replace("u+", "U+")

            if "-" in item_stripped and not item_stripped.startswith("-"):
                parts = item_stripped.split("-")
                if len(parts) != 2:
                    raise ValueError(f"Invalid range format: {item_stripped}")

                start = int(parts[0].replace("U+", ""), 16)
                end = int(parts[1].replace("U+", ""), 16)

                if start > end:
                    raise ValueError(
                        f"Invalid range: start ({hex(start)}) > end ({hex(end)})"
                    )

                codepoints.update(range(start, end + 1))
            else:
                codepoint = int(item_stripped.replace("U+", ""), 16)
                codepoints.add(codepoint)

        elif isinstance(item, int):
            if item < 0:
                raise ValueError(f"Invalid codepoint: {item} (must be non-negative)")
            codepoints.add(item)

        else:
            raise ValueError(
                f"Unsupported unicode_range item type: {type(item).__name__}"
            )

    return codepoints


class FloatTolerance:
    """Utility class for floating point math with a defined tolerance."""

    TOLERANCE = 1e-9

    @classmethod
    def equal(cls, a: float, b: float) -> bool:
        """Check if two floats are almost equal within the defined tolerance."""
        return abs(a - b) <= cls.TOLERANCE

    @classmethod
    def not_equal(cls, a: float, b: float) -> bool:
        """Check if two floats are not almost equal within the defined tolerance."""
        return not cls.equal(a, b)

    @classmethod
    def is_zero(cls, a: float) -> bool:
        """Check if a float is almost zero within the defined tolerance."""
        return abs(a) <= cls.TOLERANCE

    @classmethod
    def less_than(cls, a: float, b: float) -> bool:
        """Check if a is less than b considering the defined tolerance."""
        return (b - a) > cls.TOLERANCE

    @classmethod
    def greater_than(cls, a: float, b: float) -> bool:
        """Check if a is greater than b considering the defined tolerance."""
        return (a - b) > cls.TOLERANCE

    @classmethod
    def less_equal(cls, a: float, b: float) -> bool:
        """Check if a is less than or almost equal to b considering the defined tolerance."""
        return cls.less_than(a, b) or cls.equal(a, b)

    @classmethod
    def greater_equal(cls, a: float, b: float) -> bool:
        """Check if a is greater than or almost equal to b considering the defined tolerance."""
        return cls.greater_than(a, b) or cls.equal(a, b)



from pathlib import Path

import fpdf
import pytest
from fpdf.errors import FPDFException, FPDFUnicodeEncodingException
from fpdf.image_parsing import get_img_info
from PIL import Image

from test.conftest import assert_same_file

HERE = Path(__file__).resolve().parent


def test_add_page_throws_without_page():
    pdf = fpdf.FPDF()
    with pytest.raises(FPDFException) as e:
        pdf.text(1, 2, "ok")

    msg = "No page open, you need to call add_page() first"
    assert str(e.value) == msg


def test_encoding_exception():
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=15)
    with pytest.raises(FPDFUnicodeEncodingException) as error:
        pdf.cell(text="Joséō")
        # This should through an error since Helvetica is a latin-1 encoder and the ō is out of range.
    msg = (
        'Character "ō" at index 4 in text is outside the range of characters supported by the font '
        'used: "helvetica". Please consider using a Unicode font.'
    )
    assert str(error.value) == msg


def test_orientation_portrait_landscape():
    l = fpdf.FPDF(orientation="l")
    landscape = fpdf.FPDF(orientation="landscape")
    p = fpdf.FPDF(orientation="p")
    portrait = fpdf.FPDF(orientation="portrait")

    assert l.w_pt == landscape.w_pt
    assert l.h_pt == landscape.h_pt
    assert l.def_orientation == landscape.def_orientation

    assert p.w_pt == portrait.w_pt
    assert p.h_pt == portrait.h_pt
    assert p.def_orientation == portrait.def_orientation

    assert landscape.w_pt == portrait.h_pt
    assert landscape.h_pt == portrait.w_pt


def test_incorrect_orientation():
    with pytest.raises(ValueError) as error:
        fpdf.FPDF(orientation="hello")

    assert str(error.value) == "HELLO is not a valid PageOrientation"


def test_units():
    with pytest.raises(ValueError) as e:
        fpdf.FPDF(unit="smiles")

    assert str(e.value) == "Incorrect unit: smiles"

    assert fpdf.FPDF(unit="pt").k == pytest.approx(1)
    assert fpdf.FPDF(unit="mm").k == pytest.approx(2.83464566929)
    assert fpdf.FPDF(unit="cm").k == pytest.approx(28.3464566929)
    assert fpdf.FPDF(unit="in").k == pytest.approx(72)


def test_doc_option_only_core_fonts_encoding():
    pdf = fpdf.FPDF()
    with pytest.warns(DeprecationWarning) as record:
        pdf.set_doc_option("core_fonts_encoding", 4)
        assert pdf.core_fonts_encoding == 4

        with pytest.raises(FPDFException) as e:
            pdf.set_doc_option("not core_fonts_encoding", None)

        msg = 'Unknown document option "not core_fonts_encoding"'
        assert str(e.value) == msg

    for r in record:
        assert_same_file(r.filename, __file__)


def test_adding_content_after_closing():
    pdf = fpdf.FPDF()
    pdf.set_font("helvetica", size=24)
    pdf.add_page()
    pdf.cell(w=pdf.epw, text="Hello fpdf2!", align="C")
    pdf.output()
    with pytest.raises(FPDFException) as error:
        pdf.add_page()
    assert (
        str(error.value)
        == "A page cannot be added on a closed document, after calling output()"
    )
    with pytest.raises(FPDFException) as error:
        pdf.cell(w=pdf.epw, text="Hello again!", align="C")
    assert (
        str(error.value)
        == "Content cannot be added on a finalized document, after calling output()"
    )


def test_unsupported_image_filter_error():
    image_filter = "N/A"
    with pytest.raises(FPDFException) as error:
        get_img_info(
            HERE / "flowers.png",
            Image.open(HERE / "flowers.png"),
            image_filter=image_filter,
        )
    assert str(error.value) == f'Unsupported image filter: "{image_filter}"'


def test_incorrect_number_of_pages_toc():
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.insert_toc_placeholder(lambda a, b: None, 10)
    with pytest.raises(FPDFException):
        pdf.output()


def test_invalid_page_background():
    pdf = fpdf.FPDF()
    i = 0
    with pytest.raises(TypeError) as error:
        pdf.set_page_background(i)

    msg = f"""background must be of type str, io.BytesIO, PIL.Image.Image, drawing.DeviceRGB, tuple or None
        got: {type(i)}"""
    assert str(error.value) == msg


def test_instantiating_fpdf_module():  # issue 683
    # pylint: disable=import-outside-toplevel,not-callable,redefined-outer-name
    from fpdf import fpdf

    with pytest.raises(TypeError) as error:
        fpdf()
    assert str(error.value) == (
        "You tried to instantied the fpdf module."
        " You probably want to import the FPDF class instead:"
        " from fpdf import FPDF"
    )



from pathlib import Path
import lzma

from fpdf import FPDF
from fpdf.bidi import BidiParagraph, auto_detect_base_direction
from fpdf.enums import TextDirection
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE / ".." / "fonts"

# The BidiTest.txt file lists the input string as a list of character classes.
# CHAR_MAPPING is mapping one character of each class so we can test the algorithm with a string.
# i.e. translate L ET CS ON to "a$,!"
CHAR_MAPPING = {
    "AL": "\u0608",  # Arabic Letter
    "AN": "\u0600",  # Arabic Number
    "B": "\u2029",  # Paragraph Separator
    "BN": "\U000e007a",  # Boundary Neutral
    "CS": ",",  # Common Separator
    "EN": "0",  # European Number
    "ES": "+",  # European Separator
    "ET": "$",  # European Terminator
    "FSI": "\u2068",  # First Strong Isolate
    "L": "a",  # Left-to-right
    "NSM": "\u0303",  # Nonspacing Mark
    "R": "\u05be",  # Right-to-left
    "LRE": "\u202a",  # Left-to-Right Embedding
    "LRI": "\u2066",  # Left-to-Right Isolate
    "LRO": "\u202d",  # Left-to-Right Override
    "ON": "!",  # Other Neutrals
    "PDF": "\u202c",  # Pop Directional Format
    "PDI": "\u2069",  # Pop Directional Isolate
    "RLE": "\u202b",  # Right-to-Left Embedding
    "RLI": "\u2067",  # Right-to-Left Isolate
    "RLO": "\u202e",  # Right-to-Left Override
    "S": "\t",  # Segment Separator
    "WS": " ",  # White Space
}


def test_bidi_conformance():
    """
    The file BidiTest.txt comprises exhaustive test sequences of bidirectional types
    https://www.unicode.org/reports/tr41/tr41-32.html#Tests9
    This file contains 770,241 tests

    The file BidiTest.txt is available from:
    https://www.unicode.org/Public/17.0.0/ucd/BidiTest.txt

    The local copy is compressed with xz to save space (the original file is 7.7 MB)
    """

    def check_result(string, base_direction, levels, reorder):
        characters, reordered_characters = BidiParagraph(
            text=string, base_direction=base_direction
        ).get_all()
        levels = [i for i in levels if i != "x"]
        if not levels:
            len_levels = 0
        else:
            len_levels = len(levels)
        if len(characters) != len_levels:
            return False
        for index, char in enumerate(characters):
            if levels[index] != "x" and levels[index] != str(char.embedding_level):
                return False
        return not any(
            reorder[index] != str(char.character_index)
            for (index, char) in enumerate(reordered_characters)
        )

    with lzma.open(HERE / "BidiTest.txt.xz", "rb") as compressed_file:
        data = compressed_file.read().decode("utf-8").split("\n")

    levels = []
    reorder = []
    test_count = 0
    for line in data:
        if len(line) == 0:  # ignore blank line
            continue
        if line[0] == "#":  # ignore comment line
            continue
        if line.startswith("@Levels:"):
            levels = line[9:].split(" ")
            continue
        if line.startswith("@Reorder:"):
            reorder = line[10:].split(" ")
            continue
        test_data = line.split(";")
        assert len(test_data) == 2

        string = ""
        for char_type in test_data[0].split(" "):
            string += CHAR_MAPPING[char_type]

        bitset = test_data[1]
        if int(bitset) & 1 > 0:  # auto-detect
            assert check_result(string, None, levels, reorder)
            test_count += 1
        if int(bitset) & 2 > 0:  # force LTR
            assert check_result(string, TextDirection.LTR, levels, reorder)
            test_count += 1
        if int(bitset) & 4 > 0:  # force RTL
            assert check_result(string, TextDirection.RTL, levels, reorder)
            test_count += 1
    assert test_count == 770241


def test_bidi_character():
    """
    The other test file, BidiCharacterTest.txt, contains test sequences of explicit code points, including, for example, bracket pairs.
    There are 91,707 tests on this file

    The file BidiCharacterTest.txt is available from:
    https://www.unicode.org/Public/17.0.0/ucd/BidiCharacterTest.txt

    The local copy is compressed with xz to save space (the original file is 6.7 MB)
    """

    with lzma.open(HERE / "BidiCharacterTest.txt.xz", "rb") as compressed_file:
        data = compressed_file.read().decode("utf-8").split("\n")

    test_count = 0
    for line in data:
        if len(line) == 0:  # ignore blank line
            continue
        if line[0] == "#":  # ignore comment line
            continue
        test_data = line.split(";")
        assert len(test_data) == 5
        test_count += 1

        string = ""
        for char in test_data[0].split(" "):
            string += chr(int(char, 16))
        if test_data[1] == "0":
            base_direction = TextDirection.LTR
        elif test_data[1] == "1":
            base_direction = TextDirection.RTL
        elif test_data[1] == "2":
            base_direction = None  # auto
        else:
            raise ValueError(f"Invalid base direction {test_data[1]}")

        if not base_direction:
            # test the auto detect direction algorithm
            assert (
                auto_detect_base_direction(string) == TextDirection.LTR
                and test_data[2] == "0"
            ) or (
                auto_detect_base_direction(string) == TextDirection.RTL
                and test_data[2] == "1"
            )

        characters = BidiParagraph(
            text=string, base_direction=base_direction
        ).get_characters_with_embedding_level()

        result_index = 0
        for level in test_data[3].split(" "):
            if level == "x":
                continue
            assert int(level) == characters[result_index].embedding_level
            result_index += 1

    assert test_count == 91707


def test_bidi_lorem_ipsum(tmp_path):
    # taken from https://ar.lipsum.com/ - contains Arabic (RTL) with portions in english (LTR) within the text
    ARABIC_LOREM_IPSUM = """
ما فائدته ؟
هناك حقيقة مثبتة منذ زمن طويل وهي أن المحتوى المقروء لصفحة ما سيلهي القارئ عن التركيز على الشكل الخارجي للنص أو شكل توضع الفقرات في الصفحة التي يقرأها. ولذلك يتم استخدام طريقة لوريم إيبسوم لأنها تعطي توزيعاَ طبيعياَ -إلى حد ما- للأحرف عوضاً عن استخدام "هنا يوجد محتوى نصي، هنا يوجد محتوى نصي" فتجعلها تبدو (أي الأحرف) وكأنها نص مقروء. العديد من برامح النشر المكتبي وبرامح تحرير صفحات الويب تستخدم لوريم إيبسوم بشكل إفتراضي كنموذج عن النص، وإذا قمت بإدخال "lorem ipsum" في أي محرك بحث ستظهر العديد من المواقع الحديثة العهد في نتائج البحث. على مدى السنين ظهرت نسخ جديدة ومختلفة من نص لوريم إيبسوم، أحياناً عن طريق الصدفة، وأحياناً عن عمد كإدخال بعض العبارات الفكاهية إليها.

ما أصله ؟
خلافاَ للإعتقاد السائد فإن لوريم إيبسوم ليس نصاَ عشوائياً، بل إن له جذور في الأدب اللاتيني الكلاسيكي منذ العام 45 قبل الميلاد، مما يجعله أكثر من 2000 عام في القدم. قام البروفيسور "ريتشارد ماك لينتوك" (Richard McClintock) وهو بروفيسور اللغة اللاتينية في جامعة هامبدن-سيدني في فيرجينيا بالبحث عن أصول كلمة لاتينية غامضة في نص لوريم إيبسوم وهي "consectetur"، وخلال تتبعه لهذه الكلمة في الأدب اللاتيني اكتشف المصدر الغير قابل للشك. فلقد اتضح أن كلمات نص لوريم إيبسوم تأتي من الأقسام 1.10.32 و 1.10.33 من كتاب "حول أقاصي الخير والشر" (de Finibus Bonorum et Malorum) للمفكر شيشيرون (Cicero) والذي كتبه في عام 45 قبل الميلاد. هذا الكتاب هو بمثابة مقالة علمية مطولة في نظرية الأخلاق، وكان له شعبية كبيرة في عصر النهضة. السطر الأول من لوريم إيبسوم "Lorem ipsum dolor sit amet.." يأتي من سطر في القسم 1.20.32 من هذا الكتاب.

للمهتمين قمنا بوضع نص لوريم إبسوم القياسي والمُستخدم منذ القرن الخامس عشر في الأسفل. وتم أيضاً توفير الأقسام 1.10.32 و 1.10.33 من "حول أقاصي الخير والشر" (de Finibus Bonorum et Malorum) لمؤلفه شيشيرون (Cicero) بصيغها الأصلية، مرفقة بالنسخ الإنكليزية لها والتي قام بترجمتها هـ.راكهام (H. Rackham) في عام 1914.

"""
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="NotoArabic", fname=HERE / "NotoNaskhArabic-Regular.ttf")
    pdf.add_font(family="NotoSans", fname=FONTS_DIR / "NotoSans-Regular.ttf")
    pdf.set_fallback_fonts(["NotoSans"])
    pdf.set_font("NotoArabic", size=20)
    pdf.set_text_shaping(True)
    pdf.multi_cell(
        text=ARABIC_LOREM_IPSUM, w=pdf.epw, h=10, new_x="LEFT", new_y="NEXT", align="R"
    )

    assert_pdf_equal(
        pdf,
        HERE / "bidi_arabic_lorem_ipsum.pdf",
        tmp_path,
    )


def test_bidi_paragraph_direction(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="SBL_Hbrw", fname=HERE / "SBL_Hbrw.ttf")
    pdf.set_font("SBL_Hbrw", size=18)
    pdf.set_text_shaping(True)

    text1 = "אנגלית (באנגלית: English)"  # first char is RTL
    text2 = "The test is: אנגלית (באנגלית: English)"  # first char is LTR

    pdf.cell(text="No text shaping (not bidirectional)", new_x="left", new_y="next")
    pdf.cell(text=text1, new_x="left", new_y="next")
    pdf.cell(text=text2, new_x="left", new_y="next")
    pdf.ln()

    pdf.set_text_shaping(use_shaping_engine=True)
    pdf.cell(
        text="Text shaping-Automatic paragraph direction", new_x="left", new_y="next"
    )
    pdf.cell(text=text1, new_x="left", new_y="next")
    pdf.cell(text=text2, new_x="left", new_y="next")
    pdf.ln()

    pdf.set_text_shaping(use_shaping_engine=True, direction="rtl")
    pdf.cell(
        text="Text shaping-Force paragraph direction RTL", new_x="left", new_y="next"
    )
    pdf.cell(text=text1, new_x="left", new_y="next")
    pdf.cell(text=text2, new_x="left", new_y="next")
    pdf.ln()

    pdf.set_text_shaping(use_shaping_engine=True, direction="ltr")
    pdf.cell(
        text="Text shaping-Force paragraph direction LTR", new_x="left", new_y="next"
    )
    pdf.cell(text=text1, new_x="left", new_y="next")
    pdf.cell(text=text2, new_x="left", new_y="next")
    pdf.ln()

    assert_pdf_equal(
        pdf,
        HERE / "bidi_paragraph_direction.pdf",
        tmp_path,
    )


def test_bidi_get_string_width(tmp_path):
    # issue 1231
    teststrings = [
        "الملوك",
        "الملوك",
        "test",
        "الملوك",
        "الملوك",
        "الملوك",
        "test",
        "الملوك",
        "test",
    ]
    pdf = FPDF()
    pdf.add_page()
    pdf.c_margin = 0
    pdf.add_font("noto", style="", fname=HERE / "NotoNaskhArabic-Regular.ttf")
    pdf.set_text_shaping(use_shaping_engine=True)

    pdf.set_font("noto", size=24)
    pdf.set_draw_color(160)
    pdf.set_line_width(0.3)
    for string in teststrings:
        pdf.set_x(110 - pdf.get_string_width(string))
        pdf.rect(
            pdf.get_x(), pdf.get_y() + 2, pdf.get_string_width(string), 13, style="D"
        )
        pdf.cell(h=17, text=string)
        pdf.ln()
    pdf.ln()
    assert_pdf_equal(pdf, HERE / "bidi_get_string_width.pdf", tmp_path)



import pytest

from pathlib import Path

from fpdf import FPDF

HERE = Path(__file__).resolve().parent


def test_TitleStyle_deprecation():
    # pylint: disable=import-outside-toplevel
    with pytest.warns(DeprecationWarning):
        from fpdf import TitleStyle

        TitleStyle()

    with pytest.warns(DeprecationWarning):
        from fpdf.fonts import TitleStyle

        TitleStyle()


def test_add_font_uni_deprecation():
    pdf = FPDF()

    with pytest.warns(DeprecationWarning) as record:
        # pylint: disable=unexpected-keyword-arg
        pdf.add_font(
            "DejaVu",
            "",
            HERE.parent / "fonts" / "DejaVuSans.ttf",
            uni=True,
        )
    assert (
        str(record[0].message)
        == '"uni" parameter is deprecated since v2.5.1 and will be removed in a future release'
    )


def test_output_dest_deprecation():
    pdf = FPDF()

    with pytest.warns(DeprecationWarning) as record:
        # pylint: disable=unexpected-keyword-arg
        pdf.output(dest="S")
    assert (
        str(record[0].message)
        == '"dest" parameter is deprecated since v2.2.0 and will be removed in a future release'
    )



from filecmp import cmp

import fpdf
import pytest

from test.conftest import assert_same_file


def test_repeated_calls_to_output(tmp_path):
    pdf = fpdf.FPDF()
    pdf.output(tmp_path / "empty.pdf")
    pdf.output(tmp_path / "empty2.pdf")
    assert cmp(tmp_path / "empty.pdf", tmp_path / "empty2.pdf")


def test_deprecation_warning(tmp_path):
    pdf = fpdf.FPDF()
    with pytest.warns(DeprecationWarning) as record:
        # pylint: disable=unexpected-keyword-arg
        pdf.output(name=tmp_path / "empty.pdf", dest="F")
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)


def test_save_to_absolute_path(tmp_path):
    pdf = fpdf.FPDF()
    pdf.output((tmp_path / "empty.pdf").absolute())



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def next_row(pdf):
    pdf.ln()
    pdf.set_y(pdf.get_y() + size + margin)


size = 50
start_angle, end_angle = 30, 130
margin = 10


def test_solid_arc_not_circle(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["F", "FD", "DF", None]):
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size / 2,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            style=style,
        )
        pdf.set_x(pdf.get_x() + (size / 2) + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_not_circle.pdf", tmp_path)


def test_solid_arc_style(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["F", "FD", "DF", None]):
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            style=style,
        )
        pdf.set_x(pdf.get_x() + size + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_style.pdf", tmp_path)


def test_solid_arc_line_width(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    size_1 = size

    for line_width in [1, 2, 3]:
        pdf.set_line_width(line_width)
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            # A workaround to avoid warning of pylint triggering duplicate code
            # with test_arc.py
            a=size_1,
            b=size_1,
            start_angle=start_angle,
            end_angle=end_angle,
            style=None,
        )
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)
    for line_width in [4, 5, 6]:
        pdf.set_line_width(line_width)
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            style=None,
        )
        pdf.set_x(pdf.get_x() + size + margin)
    pdf.set_line_width(0.2)  # reset

    assert_pdf_equal(pdf, HERE / "class_solid_arc_line_width.pdf", tmp_path)


def test_solid_arc_draw_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pdf.set_line_width(0.5)
    for gray in [70, 140, 210]:
        pdf.set_draw_color(gray)
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            style=None,
        )
        pdf.set_x(pdf.get_x() + size + margin)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_draw_color.pdf", tmp_path)


def test_solid_arc_fill_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pdf.set_fill_color(240)
    for color in [[230, 30, 180], [30, 180, 30], [30, 30, 70]]:
        pdf.set_draw_color(*color)
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            style="FD",
        )
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_fill_color.pdf", tmp_path)


def test_solid_arc_inclination(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    size_1 = size

    for counter, inclination in enumerate([0, 30, 90, 120]):
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            # A workaround to avoid warning of pylint triggering duplicate code
            # with test_arc.py
            a=size_1,
            b=size_1,
            start_angle=start_angle,
            end_angle=end_angle,
            inclination=inclination,
            style=None,
        )
        pdf.set_x(pdf.get_x() + size + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_inclination.pdf", tmp_path)


def test_arc_clockwise(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for inclination in [0, 30, 90, 120]:
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            inclination=inclination,
            style=None,
        )
        pdf.set_x(pdf.get_x() + size + margin)
        pdf.solid_arc(
            x=pdf.get_x(),
            y=pdf.get_y(),
            a=size,
            b=size,
            start_angle=start_angle,
            end_angle=end_angle,
            inclination=inclination,
            clockwise=True,
            style=None,
        )
        next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_solid_arc_clockwise.pdf", tmp_path)



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

import pytest

HERE = Path(__file__).resolve().parent


def test_star(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    # no fill with line test
    y = 20
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="D")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="D")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="D")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="D")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="D")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="D")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="D")

    # fill and color test
    y += 40
    pdf.set_fill_color(r=134, g=200, b=15)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="DF")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="DF")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="DF")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="DF")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="DF")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="DF")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="DF")

    # draw color test
    y += 40
    pdf.set_fill_color(r=0, g=0, b=0)
    pdf.set_draw_color(r=204, g=0, b=204)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="D")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="D")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="D")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="D")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="D")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="D")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="D")

    # line width test
    y += 40
    pdf.set_line_width(1)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="D")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="D")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="D")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="D")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="D")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="D")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="D")

    # line color and fill color
    y += 40
    pdf.set_fill_color(r=3, g=190, b=252)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="DF")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="DF")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="DF")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="DF")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="DF")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="DF")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="DF")

    # fill only
    y += 40
    pdf.set_draw_color(r=0, g=0, b=255)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, style="F")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, style="F")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, style="F")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, style="F")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, style="F")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, style="F")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, style="F")

    # rotation test
    y += 40
    pdf.set_draw_color(r=0, g=0, b=255)
    pdf.star(x=15, y=y, r_in=5, r_out=15, corners=3, rotate_degrees=0, style="DF")
    pdf.star(x=45, y=y, r_in=5, r_out=15, corners=4, rotate_degrees=35, style="DF")
    pdf.star(x=75, y=y, r_in=5, r_out=15, corners=5, rotate_degrees=45, style="DF")
    pdf.star(x=105, y=y, r_in=5, r_out=15, corners=6, rotate_degrees=200, style="DF")
    pdf.star(x=135, y=y, r_in=5, r_out=15, corners=7, rotate_degrees=13, style="DF")
    pdf.star(x=165, y=y, r_in=5, r_out=15, corners=8, rotate_degrees=22.5, style="DF")
    pdf.star(x=195, y=y, r_in=5, r_out=15, corners=9, rotate_degrees=77.3, style="DF")

    assert_pdf_equal(pdf, HERE / "regular_star.pdf", tmp_path)


def test_star_invalid_style():
    pdf = fpdf.FPDF()
    pdf.add_page()

    with pytest.raises(ValueError):
        pdf.star(x=15, y=15, r_in=5, r_out=15, corners=3, rotate_degrees=0, style="N")



#!/usr/bin/env python3
from itertools import chain
from string import digits, ascii_uppercase, ascii_lowercase
from fpdf import FPDF

pdf = FPDF()
pdf.add_page()
pdf.set_font("helvetica", size=30)
pdf.cell(w=pdf.epw, text="ZapfDingbats", align="C")
pdf.set_margin(0)
for i, char in enumerate(
    "!\"#$%&'()*+,-./"
    + digits
    + ":;<=>?@"
    + ascii_uppercase
    + "[\\]^_`"
    + ascii_lowercase
    + "{|}~"
):
    if i % 24 == 0:
        pdf.x = pdf.l_margin = 10 + (i // 24) * 50
        pdf.y = 30
    pdf.set_font("helvetica", size=30)
    pdf.cell(text=char + " = ")
    pdf.set_font("zapfdingbats", size=30)
    pdf.cell(text=char, new_x="LMARGIN", new_y="NEXT")
for i, n in enumerate(chain(range(0x80, 0x8E), range(0xA1, 0xF0), range(0xF1, 0xFF))):
    if i % 26 == 0:
        col = i // 26
        if col % 3 == 0:
            pdf.add_page()
        pdf.x = pdf.l_margin = 10 + (col % 3) * 65
        pdf.y = 10
    pdf.set_font("helvetica", size=30)
    pdf.cell(text=f"\\u00{hex(n)[2:]} = ")
    pdf.set_font("zapfdingbats", size=30)
    pdf.cell(text=chr(n), new_x="LMARGIN", new_y="NEXT")
pdf.output("zapfdingbats.pdf")
